"""
GUI tools using tkinter. The multicolumn list is inspired by a post on
Stackoverflow:

http://stackoverflow.com/questions/5286093/\
display-listbox-with-columns-using-tkinter
"""
import os
import tkinter as tk
import tkinter.filedialog as filedialog
import tkinter.font as tkFont
import tkinter.ttk as ttk

LASTOPENDIR = None
LASTSAVEDIR = None

__all__ = [
    "get_file_name",
    "askopenfilename",
    "asksaveasfilename",
    "MultiColumnListbox",
]


def get_file_name(f, read):
    """
    Utility function to get the file name using guitools if needed.

    Parameters
    ----------
    f : string or None
        File name or directory name or None. If `f` is not None or is
        not a directory name, this routine just returns it as is. If
        it is a directory name, it is used as the `initialdir` option
        to :func:`askopenfilename` or :func:`asksaveasfilename`.
    read : bool
        If True, calls :func:`askopenfilename` to get file name.
        Otherwise, :func:`asksaveasfilename` is called.

    Returns
    -------
    filename : string
        The selected filename.
    """
    if isinstance(f, str) and os.path.isdir(f):  # pragma: no cover
        initialdir = f
        f = None
    else:
        initialdir = None
    if f is None:  # pragma: no cover
        if read:
            return askopenfilename(initialdir=initialdir)
        else:
            return asksaveasfilename(initialdir=initialdir)
    return f


def askopenfilename(title=None, filetypes=None, initialdir=None):  # pragma: no cover
    """
    Use GUI to select file for reading

    Parameters
    ----------
    title : string or None
        Title of window. Use None to accept tkinter default.
    filetypes : list or None
        Option to limit file search to certain patterns. Typically,
        this is a list of tuples, each tuple containing a description
        followed by a pattern (see example below).
    initialdir : string or None
        Initial directory to begin search. If None, use previous
        location if there is one.

    Returns
    -------
    filename : string
        The selected filename.

    Notes
    -----
    Here is a simple example::

        from pyyeti import guitools
        filename = guitools.askopenfilename()

    To filter the files to selected types::

        from pyyeti import guitools
        filetypes = [('Output4 files', '*.op4'),
                     ('All files', '*')]
        filename = guitools.askopenfilename(filetypes=filetypes)

    """
    global LASTOPENDIR
    root = tk.Tk()
    root.withdraw()
    if filetypes is None:
        filetypes = []
    if initialdir is None:
        initialdir = LASTOPENDIR
    filename = filedialog.askopenfilename(
        parent=root, filetypes=filetypes, initialdir=initialdir, title=title
    )
    root.destroy()
    if filename:
        LASTOPENDIR = os.path.dirname(filename)
    return filename


def asksaveasfilename(title=None, filetypes=None, initialdir=None):  # pragma: no cover
    """
    Use GUI to select file for writing

    Parameters
    ----------
    title : string or None
        Title of window. Use None to accept tkinter default.
    filetypes : list or None
        Option to limit file search to certain patterns. Typically,
        this is a list of tuples, each tuple containing a description
        followed by a pattern (see example below).
    initialdir : string or None
        Initial directory to begin search. If None, use previous
        location if there is one.

    Returns
    -------
    filename : string
        The selected filename.

    Notes
    -----
    Here is a simple example::

        from pyyeti import guitools
        filename = guitools.asksaveasfilename()

    To filter the files to selected types::

        from pyyeti import guitools
        filetypes = [('Output4 files', '*.op4'),
                     ('All files', '*')]
        filename = guitools.asksaveasfilename(filetypes=filetypes)

    """
    global LASTSAVEDIR
    root = tk.Tk()
    root.withdraw()
    if filetypes is None:
        filetypes = []
    if initialdir is None:
        initialdir = LASTSAVEDIR
    filename = filedialog.asksaveasfilename(
        parent=root, filetypes=filetypes, initialdir=initialdir, title=title
    )
    # self.root.after_idle(self._quit)
    root.destroy()
    if filename:
        LASTSAVEDIR = os.path.dirname(filename)
    return filename


class MultiColumnListbox(object):  # pragma: no cover
    """
    Use a ttk.TreeView to build a linked, multicolumn listbox.

    Once a window is created, you may select a single item by either
    selecting a row and pressing "Done" or by double-clicking a row.
    Or, you may select multiple items and pressing "Done".

    You can filter the values shown by entering strings in the filter
    boxes above the multicolumn listbox and hitting the Return
    key. All values in a row must match their respective filter for
    the row to remain visible. A value matches if it contains the
    filter anywhere in it.

    After selection, retrieve your selection by accessing attributes
    `sel_index` or `sel_dict`. `sel_index` is a list of indexes into
    the provided lists. `sel_dict` is a dictionary with the keys being
    the indexes and the values being another dictionary of `header` :
    list-item pairs.

    Example code::

        from pyyeti import guitools
        headers = ['First', 'Middle', 'Last']
        lst1 = ['Tony', 'Jennifer', 'Albert', 'Marion']
        lst2 = ['J.', 'M.', 'E.', 'K.']
        lst3 = ['Anderson', 'Smith', 'Kingsley', 'Cotter']
        ind = guitools.MultiColumnListbox(
                 'Select person', headers, [lst1, lst2, lst3]
              ).sel_index[0]
        print(f'First Person is {lst1[ind]} {lst2[ind]} {lst3[ind]}')

    Or, using the `sel_dict` attribute::

        dct = guitools.MultiColumnListbox(
                 'Select person', headers, [lst1, lst2, lst3]
              ).sel_dict
        key = sorted(dct)[0]
        vals = dct[key]
        print(f"First Person is {vals['First']} {vals['Middle']} "
              f"{vals['Last']}")

    """

    def __init__(
        self,
        title,
        headers,
        lists,
        topstring=(
            "Click on header to sort by that column;\n"
            "Drag boundary to change width of column"
        ),
    ):
        """
        Initialize a :class:`MultiColumnListbox` instance.

        Parameters
        ----------
        title : string
            Title for the window
        headers : list of strings
            List of column headers
        lists : list of lists
            Corresponds to `headers`. Each list must be the same
            length and contain the contents of the columns. The
            contents are expected to be strings.
        topstring : string; optional
            String to print above table
        """
        self.root = tk.Tk()
        self.root.title(title)
        self.tree = None
        self.headers = headers
        self.lists = lists
        self.topstring = topstring
        self.sel_dict = {}
        self.sel_index = []
        self.detached_items = [0] * len(lists[0])
        self._setup_widgets()
        self._build_tree()
        self.root.mainloop()

    def _setup_widgets(self):
        msg = tk.Text(wrap="word", height=2, font="TkDefaultFont")
        msg.insert("1.0", self.topstring)  # line 1, column 0
        msg.configure(bg=self.root.cget("bg"), relief="flat", state="disabled")
        msg.pack(fill="x")
        self._add_filter_boxes()
        container = ttk.Frame()
        container.pack(fill="both", expand=True)
        # create a treeview with dual scrollbars
        self.tree = ttk.Treeview(
            height=min(25, len(self.lists[0])), columns=self.headers, show="headings"
        )
        vsb = ttk.Scrollbar(orient="vertical", command=self.tree.yview)
        hsb = ttk.Scrollbar(orient="horizontal", command=self.tree.xview)
        self.tree.configure(yscrollcommand=vsb.set, xscrollcommand=hsb.set)
        self.tree.grid(column=0, row=0, sticky="nsew", in_=container)
        vsb.grid(column=1, row=0, sticky="ns", in_=container)
        hsb.grid(column=0, row=1, sticky="ew", in_=container)
        container.grid_columnconfigure(0, weight=1)
        container.grid_rowconfigure(0, weight=1)
        # add a done button:
        button = ttk.Button(text="Done", command=self._get_selection)
        button.pack()

    def _add_filter_boxes(self):
        container = ttk.Frame()
        container.pack(fill="x")
        msg = ttk.Label(
            container,
            text="Filters:",
            # background='gray',
            borderwidth=5,
        )
        # msg.grid(row=0, rowspan=2, column=0, padx=10)
        msg.grid(row=0, column=0, padx=10)

        self.casesen = tk.IntVar()
        self.casesen.set(0)
        cbutton = ttk.Checkbutton(
            container, text="Case Sensitive", variable=self.casesen
        )  # command=self._checkbutton_action)
        cbutton.grid(row=1, column=0)

        self.filter_var = []
        # add an Entry widget for all columns except "Index":
        for i, header in enumerate(self.headers):
            filter_var = tk.StringVar()
            filter_entry = ttk.Entry(container, textvariable=filter_var)
            filter_var.set("")
            filter_entry.bind("<Key-Return>", self._apply_filters)
            # filter_entry.pack(side='left', padx=10, expand=True)
            ttk.Label(container, text=header).grid(row=0, column=i + 1, sticky="w")
            filter_entry.grid(row=1, column=i + 1, sticky="w")
            # container.grid_columnconfigure(i+1, weight=1)
            self.filter_var.append(filter_var)

    def _apply_filters(self, event=None):
        # detach all items remaining:
        for item in self.tree.get_children(""):
            i = int(item[1:], 16) - 1
            self.detached_items[i] = item
            self.tree.detach(item)
        # reattach those where all filters pass:
        if self.casesen.get():
            filtervars = [v.get() for v in self.filter_var]

            def _do_find(value, sub):
                return value.find(sub) > -1

        else:
            filtervars = [v.get().lower() for v in self.filter_var]

            def _do_find(value, sub):
                return value.lower().find(sub) > -1

        for i in range(len(self.lists[0])):
            for string, curlist in zip(filtervars, self.lists):
                if string and not _do_find(curlist[i], string):
                    break
            else:
                # only here if the 'break' was not executed ...
                # when all filters pass:
                self.tree.move(self.detached_items[i], "", i)
                self.detached_items[i] = 0

    def _quit(self):
        # self.root.quit()
        self.root.destroy()

    def _store_selection(self, items):
        dct = {}
        index = []
        for item in items:
            i = int(item[1:], 16) - 1
            dct[i] = self.tree.set(item)
            index.append(i)
        self.sel_dict = dct
        self.sel_index = index
        self.root.after_idle(self._quit)

    def _get_selection(self):
        self._store_selection(self.tree.selection())

    def _double_click(self, event):
        item = self.tree.identify("item", event.x, event.y)
        self._store_selection((item,))

    def _build_tree(self):
        for col in self.headers:
            self.tree.heading(
                col, text=col.title(), command=lambda c=col: _sortby(self.tree, c, 0)
            )
            # adjust the column's width to the header string
            # - add 15 pixels for a little buffer
            self.tree.column(col, width=tkFont.Font().measure(col.title()) + 15)

        for item in zip(*self.lists):
            self.tree.insert("", "end", values=item)

        # adjust each column's width by maximum length string:
        for i, col in enumerate(self.headers):
            try:
                s = max(self.lists[i], key=len)
            except TypeError:
                s = str(max(self.lists[i]))
            col_w = tkFont.Font().measure(s)
            width = self.tree.column(col, width=None)
            if width < col_w:
                self.tree.column(col, width=col_w)
        self.tree.bind("<Double-1>", self._double_click)


def _sortby(tree, col, descending):  # pragma: no cover
    """sort tree contents when a column header is clicked on"""
    # grab values to sort
    data = [(tree.set(child, col), child) for child in tree.get_children("")]

    # print(data)
    data.sort(reverse=descending)
    for ix, item in enumerate(data):
        tree.move(item[1], "", ix)

    # to make first of any current selection visible:
    tree.see(tree.selection()[:1])

    # switch the heading so it will sort in the opposite direction
    tree.heading(col, command=lambda col=col: _sortby(tree, col, int(not descending)))



# -*- coding: utf-8 -*-
"""
Low level tool for writing percent difference reports. Typically, this
is called via: :func:`cla.DR_Results.rptpct`.
"""
from io import StringIO
from types import SimpleNamespace
import warnings
import numpy as np
import matplotlib.pyplot as plt
from pyyeti import ytools, locate, writer
from ._utilities import _get_rpt_headers, _get_numform, _proc_filterval
from ._magpct import magpct


__all__ = ["rptpct1"]


# FIXME: We need the str/repr formatting used in Numpy < 1.14.
try:
    np.set_printoptions(legacy="1.13")
except TypeError:
    pass


def _apply_pv(value, pv, oldlen):
    # if value has a len that's > 1, try to partition it down;
    # otherwise, return it as is:
    try:
        n = len(value)
    except TypeError:
        return value
    else:
        if n == 1:
            return value

    # `value` is a vector with len > 1 ... ensure it is a true numpy
    # array:
    value = np.atleast_1d(value)

    # oldlen is either 0 (for `value` vectors that are expected to be
    # full size ... currently, only the `filterval` and
    # `magpct_filterval` vectors), or it is the length of the
    # dimension that the `value` index type of partition vector
    # (currently, only the `ignorepv` vector) was originally defined
    # to partition.
    if oldlen == 0:
        # `value` is `filterval` or `magpct_filterval` ... these just
        # need to be partitioned down:
        newvalue = value[pv]
    else:
        # `value` is `ignorepv` ... it needs to be redefined to
        # correspond to reduced size:
        truefalse = locate.index2bool(value, oldlen)
        newvalue = truefalse[pv].nonzero()[0]
    return newvalue


def _align_mxmn(mxmn1, mxmn2, labels2, row_number, infodct):
    if infodct["labels"] and infodct["labels"] != labels2:
        n = len(infodct["labels"])
        pv1, pv2 = locate.list_intersect(infodct["labels"], labels2)
        mxmn1 = mxmn1[pv1]
        mxmn2 = mxmn2[pv2]
        infodct["labels"] = [infodct["labels"][i] for i in pv1]
        row_number = row_number[pv1]
        infodct["filterval"] = _apply_pv(infodct["filterval"], pv1, 0)
        infodct["magpct_filterval"] = _apply_pv(infodct["magpct_filterval"], pv1, 0)
        infodct["ignorepv"] = _apply_pv(infodct["ignorepv"], pv1, n)
    return mxmn1, mxmn2, row_number


def _get_filtline(filterval):
    if len(filterval) > 1:
        filtline = "Filter:      <defined row-by-row>\n"
    else:
        filtline = f"Filter:      {filterval[0]}\n"
    return filtline


def _get_noteline(use_range, names, prtbads, flagbads):
    noteline = "Notes:       "
    tab = "             "
    if not use_range:
        noteline += "% Diff = +/- abs(({0}-{1})/{1})*100\n".format(*names)
    else:
        noteline += "% Diff = +/- abs({0}-{1})/max(abs({1}(max,min)))*100\n".format(
            *names
        )

    noteline += tab + "Sign set such that positive % differences indicate exceedances\n"
    prtbad, prtbadh, prtbadl = prtbads
    flagbad, flagbadh, flagbadl = flagbads
    if prtbad is not None or prtbadh is not None or prtbadl is not None:
        if prtbad is not None:
            prtbad = abs(prtbad)
            noteline += tab + f"Printing rows where abs(% Diff) > {prtbad}%\n"
        elif prtbadh is not None:
            noteline += tab + f"Printing rows where % Diff > {prtbadh}%\n"
        else:
            noteline += tab + f"Printing rows where % Diff < {prtbadl}%\n"

    if flagbad is not None or flagbadh is not None or flagbadl is not None:
        if flagbad is not None:
            flagbad = abs(flagbad)
            noteline += tab + f"Flagging (*) rows where abs(% Diff) > {flagbad}%\n"
        elif flagbadh is not None:
            noteline += tab + f"Flagging (*) rows where % Diff > {flagbadh}%\n"
        else:
            noteline += tab + f"Flagging (*) rows where % Diff < {flagbadl}%\n"
    return noteline


def _get_badpv(pct, pv, bad, badh, badl, defaultpv=False):
    if bad is not None or badh is not None or badl is not None:
        badpv = pv.copy()
        if bad is not None:
            badpv &= abs(pct) > bad
        elif badh is not None:
            badpv &= pct > badh
        else:
            badpv &= pct < badl
    else:
        badpv = np.empty(len(pct), bool)
        badpv[:] = defaultpv
    return badpv


def _get_pct_diff(a, b, filt, pv, nastring, mxmn_b=None, ismax=True, flagbads=None):
    # either can pass filter to be kept:
    pv &= (abs(a) > filt) | (abs(b) > filt)

    if mxmn_b is not None:
        denom = np.nanmax(abs(mxmn_b), axis=1)
    else:
        denom = abs(b)

    # put 1's in for filtered values ... this is temporary
    a = a.copy()
    b = b.copy()
    a[~pv] = 1.0
    b[~pv] = 1.0

    z = denom == 0.0
    denom[z] = 1.0
    pct = 100 * abs(a - b) / denom
    pct[z] = 100.0  # np.inf

    # make less extreme values negative
    neg = a < b if ismax else a > b
    pct[neg] *= -1.0

    # put nan's in for the filtered or n/a rows:
    pct[~pv] = np.nan

    # make 7 char version:
    spct = [f"{p:7.2f}" for p in pct]
    badpv = _get_badpv(pct, pv, *flagbads, False)
    for j in badpv.nonzero()[0]:
        spct[j] += "*"
    for j in (~pv).nonzero()[0]:
        spct[j] = nastring

    return pct, spct


def _get_histogram_str(desc, hdr, pctinfo):
    pctcount = pctinfo["hsto"]
    s = [
        (f"\n\n    {desc} - {hdr} Comparison Histogram\n\n"),
        ("      % Diff      Count    Percent\n     --------   --------   -------\n"),
    ]
    with StringIO() as f:
        writer.vecwrite(f, "     {:8.2f}   {:8.0f}   {:7.2f}\n", pctcount)
        s.append(f.getvalue())
        s.append("\n")

    # total_percent_10 will either be 0 or 1000:
    #  - 0 if all % diffs are "n/a"
    #  - 1000 otherwise
    total_percent_10 = np.round(pctcount[:, 2].sum() * 10)
    last = -1.0
    for pdiff in [1, 2, 5, 10, 15, 20, 25, 50, 100, 500]:
        pvdiff = abs(pctcount[:, 0]) <= pdiff
        num = pctcount[pvdiff, 2].sum()
        if num > last:
            s.append(f"    {num:.1f}% of values are within {pdiff}%\n")
        if np.round(num * 10) == total_percent_10:
            break
        last = num

    pct = pctinfo["pct"]
    n = len(pct)
    if n == 0:
        s.append(
            "\n    % Diff Statistics: [Min, Max, Mean, StdDev]"
            " = [n/a, n/a, n/a, n/a]\n"
        )
    else:
        stddev = 0.0 if n <= 1 else pct.std(ddof=1)
        s.append(
            "\n    % Diff Statistics: [Min, Max, Mean, StdDev]"
            f" = [{pct.min():.2f}, {pct.max():.2f}, {pct.mean():.4f}, {stddev:.4f}]\n"
        )
    return "".join(s)


def _proc_pct(
    ext1,
    ext2,
    filterval,
    magpct_filterval,
    *,
    names,
    mxmn1,
    comppv,
    mxmn_b,
    ismax,
    histogram_inc,
    prtbads,
    flagbads,
    numform,
    valhdr,
    maxhdr,
    minhdr,
    absmhdr,
    pdhdr,
    nastring,
    doabsmax,
    shortabsmax,
    print_info,
):
    # handle magpct stuff here:
    mag = ext1[comppv], ext2[comppv]
    if magpct_filterval is not None and len(magpct_filterval) > 1:
        magfilt = magpct_filterval[comppv]
    else:
        magfilt = magpct_filterval

    pv = comppv.copy()
    pct, spct = _get_pct_diff(
        ext1,
        ext2,
        filterval,
        pv,
        nastring,
        mxmn_b=mxmn_b,
        ismax=ismax,
        flagbads=flagbads,
    )
    pct_ret = pct[pv]
    hsto = ytools.histogram(pct_ret, histogram_inc)

    # for trimming down if prtbad set:
    prtpv = _get_badpv(pct, pv, *prtbads, True)
    pctlen = max(len(pdhdr), len(max(spct, key=len)))
    sformatpd = f"{{:{pctlen}}}"

    # for writer.formheader:
    numlen = max(13, len(max(names, key=len)), len(numform.format(np.pi)))
    if not doabsmax:
        print_info.headers1.extend([*names, ""])
        print_info.headers2.extend([valhdr, valhdr, pdhdr])
        print_info.formats.extend([numform, numform, sformatpd])
        print_info.printargs.extend([ext1, ext2, spct])
        print_info.widths.extend([numlen, numlen, pctlen])
        print_info.seps.extend([4, 2, 2])
        print_info.justs.extend(["c", "c", "c"])
    elif shortabsmax:
        print_info.headers1.extend([*names, ""])
        print_info.headers2.extend([absmhdr, absmhdr, pdhdr])
        print_info.formats.extend([numform, numform, sformatpd])
        print_info.printargs.extend([ext1, ext2, spct])
        print_info.widths.extend([numlen, numlen, pctlen])
        print_info.seps.extend([4, 2, 2])
        print_info.justs.extend(["c", "c", "c"])
    else:
        print_info.headers1.extend([names[0], names[0], names[0], names[1], ""])
        print_info.headers2.extend([maxhdr, minhdr, absmhdr, absmhdr, pdhdr])
        print_info.formats.extend([numform, numform, numform, numform, sformatpd])
        print_info.printargs.extend([mxmn1[:, 0], mxmn1[:, 1], ext1, ext2, spct])
        print_info.widths.extend([numlen, numlen, numlen, numlen, pctlen])
        print_info.seps.extend([4, 2, 2, 2, 2])
        print_info.justs.extend(["c", "c", "c", "c", "c"])
    return dict(
        pct=pct_ret, spct=spct, hsto=hsto, prtpv=prtpv, mag=mag, magfilt=magfilt
    )


def _figure_on(name, doabsmax, show_figures):
    figsize = [8.5, 11.0]
    if doabsmax:
        figsize[1] /= 3.0
    if show_figures:
        plt.figure(name, figsize=figsize)
        plt.clf()
    else:
        plt.figure(figsize=figsize)


def _figure_off(show_figures):
    if not show_figures:
        plt.close()


def _prep_subplot(pctinfo, sp):
    if "mx" in pctinfo:
        # if not just doing absmax
        if sp > 311:
            plt.subplot(sp, sharex=plt.gca())
        else:
            plt.subplot(sp)


def _plot_magpct(
    pctinfo,
    names,
    desc,
    doabsmax,
    filename,
    magpct_options,
    use_range,
    maxhdr,
    minhdr,
    absmhdr,
    show_figures,
    tight_layout_args,
):
    ptitle = f"{desc} - {{}} Comparison vs Magnitude"
    xl = f"{names[1]} Magnitude"
    yl = f"% Diff of {names[0]} vs {names[1]}"
    _figure_on("Magpct - " + desc, doabsmax, show_figures)
    try:
        for lbl, hdr, sp, ismax in (
            ("mx", maxhdr, 311, True),
            ("mn", minhdr, 312, False),
            ("amx", absmhdr, 313, True),
        ):
            _prep_subplot(pctinfo, sp)
            if lbl in pctinfo:
                if use_range:
                    ref = pctinfo["amx"]["mag"][1]
                else:
                    ref = None
                magpct(
                    pctinfo[lbl]["mag"][0],
                    pctinfo[lbl]["mag"][1],
                    Ref=ref,
                    ismax=ismax,
                    filterval=pctinfo[lbl]["magfilt"],
                    **magpct_options,
                )
                plt.title(ptitle.format(hdr))
                plt.xlabel(xl)
                plt.ylabel(yl)
            plt.grid(True)
        plt.tight_layout(**tight_layout_args)
        if isinstance(filename, str):
            plt.savefig(filename + ".magpct.png")
    finally:
        _figure_off(show_figures)


def _plot_histogram(
    pctinfo,
    names,
    desc,
    doabsmax,
    filename,
    histogram_inc,
    maxhdr,
    minhdr,
    absmhdr,
    show_figures,
    tight_layout_args,
):
    ptitle = f"{desc} - {{}} Comparison Histogram"
    xl = f"% Diff of {names[0]} vs {names[1]}"
    yl = "Percent Occurrence (%)"
    _figure_on("Histogram - " + desc, doabsmax, show_figures)
    try:
        for lbl, hdr, sp in (
            ("mx", maxhdr, 311),
            ("mn", minhdr, 312),
            ("amx", absmhdr, 313),
        ):
            _prep_subplot(pctinfo, sp)
            if lbl in pctinfo:
                width = histogram_inc
                x = pctinfo[lbl]["hsto"][:, 0]
                y = pctinfo[lbl]["hsto"][:, 2]
                colors = ["b"] * len(x)
                ax = abs(x)
                pv1 = ((ax > 5) & (ax <= 10)).nonzero()[0]
                pv2 = (ax > 10).nonzero()[0]
                for pv, c in ((pv1, "m"), (pv2, "r")):
                    for i in pv:
                        colors[i] = c
                plt.bar(x, y, width=width, color=colors, align="center")
                plt.title(ptitle.format(hdr))
                plt.xlabel(xl)
                plt.ylabel(yl)
                x = abs(max(plt.xlim(), key=abs))
                if x < 5:
                    plt.xlim(-5, 5)
            plt.grid(True)
        plt.tight_layout(**tight_layout_args)
        if isinstance(filename, str):
            plt.savefig(filename + ".histogram.png")
    finally:
        _figure_off(show_figures)


def rptpct1(
    mxmn1,
    mxmn2,
    filename,
    *,
    title="PERCENT DIFFERENCE REPORT",
    names=("Self", "Reference"),
    desc=None,
    filterval=None,
    labels=None,
    units=None,
    ignorepv=None,
    uf_reds=None,
    use_range=True,
    numform=None,
    prtbad=None,
    prtbadh=None,
    prtbadl=None,
    flagbad=None,
    flagbadh=None,
    flagbadl=None,
    dohistogram=True,
    histogram_inc=1.0,
    domagpct=True,
    magpct_options=None,
    doabsmax=False,
    shortabsmax=False,
    roundvals=-1,
    rowhdr="Row",
    deschdr="Description",
    maxhdr="Maximum",
    minhdr="Minimum",
    absmhdr="Abs-Max",
    perpage=-1,
    tight_layout_args=None,
    show_figures=False,
    align_by_label=True,
):
    """
    Write a percent difference report between 2 sets of max/min data

    Parameters
    ----------
    mxmn1 : 2d array_like or SimpleNamespace
        The max/min data to compare to the `mxmn2` set. If 2-column
        array_like, its columns are: [max, min]. If SimpleNamespace,
        it must be as defined in :class:`DR_Results` and have these
        members:

        .. code-block:: none

            .ext = [max, min]
            .drminfo = SimpleNamespace which has (at least):
               .desc      = one line description of category
               .filterval = the filter value; (see `filterval`
                            description below)
               .labels    = a list of descriptions; one per row
               .ignorepv  = these rows will get 'n/a' for % diff
               .units     = string with units
               .uf_reds   = uncertainty factors

        Note that the inputs `desc`, `labels`, etc, override the
        values above.
    mxmn2 : 2d array_like or SimpleNamespace
        The reference set of max/min data. Format is the same as
        `mxmn1`.

        .. note::
            If both `mxmn1` and `mxmn2` are SimpleNamespaces and have
            the ``.drminfo.labels`` attribute, this routine will, by
            default, use the labels to align the data sets for
            comparison. To prevent this, set the `align_by_label`
            parameter to False.

    filename : string or file_like or 1 or None
        Either a name of a file, or is a file_like object as returned
        by :func:`open` or :class:`io.StringIO`. Input as integer 1 to
        write to stdout. Can also be the name of a directory or None;
        in these cases, a GUI is opened for file selection.
    title : string; must be named; optional
        Title for the report
    names : list/tuple; must be named; optional
        Two (short) strings identifying the two sets of data
    desc : string or None; must be named; optional
        A one line description of the table. Overrides
        `mxmn1.drminfo.desc`. If neither are input,
        'No description provided' is used.
    filterval : scalar, 1d array_like or None; must be named; optional
        Numbers with absolute value <= than `filterval` will get a
        'n/a' % diff. If vector, length must match number of rows in
        `mxmn1` and `mxmn2` data. Overrides `mxmn1.drminfo.filterval`.
        If neither are input, `filterval` is set to 1.e-6.
    labels : list or None; must be named; optional
        A list of strings briefly describing each row. Overrides
        `mxmn1.drminfo.labels`. If neither are input,
        ``['Row 1','Row 2',...]`` is used.
    units : string or None; must be named; optional
        Specifies the units. Overrides `mxmn1.drminfo.units`. If
        neither are input, 'Not specified' is used.
    ignorepv : 1d array or None; must be named; optional
        0-offset index vector specifying which rows of `mxmn1` to
        ignore (they get the 'n/a' % diff). Overrides
        `mxmn1.drminfo.ignorepv`. If neither are input, no rows are
        ignored (though `filterval` is still used).

        .. note::
            `ignorepv` applies *before* any alignment by labels is
            done (when `align_by_label` is True, which is the
            default).

    uf_reds : 1d array or None; must be named; optional
        Uncertainty factors: [rigid, elastic, dynamic, static].
        Overrides `mxmn1.drminfo.uf_reds`. If neither is input,
        'Not specified' is used.
    use_range : bool; must be named, optional
        If True, the denominator of the % diff calc for both the max
        & min for each row is the absolute maximum of the reference
        max & min for that row. If False, the denominator is the
        applicable reference max or min. A quick example shows why
        ``use_range=True`` might be useful:

        .. code-block:: none

            If [max1, min1] = [12345, -10] and
               [max2, min2] = [12300,  50]
            Then:
               % diff = [0.37%,   0.49%]  if use_range is True
               % diff = [0.37%, 120.00%]  if use_range is False

        Note that the sign of the % diff is defined such that a
        positive % diff means an exceedance: where ``max1 > max2`` or
        ``min1 < min2``.

        `use_range` is ignored if `doabsmax` is True.

    numform : string or None; must be named; optional
        Format of the max & min numbers. If None, it is set internally
        to be 13 chars wide and depends on the range of numbers to
        print:

           - if range is "small", numform='{:13.xf}' where "x" ranges
             from 0 to 7
           - if range is "large", numform='{:13.6e}'

    prtbad : scalar or None; must be named; optional
        Only print rows where ``abs(%diff) > prtbad``. For example, to
        print rows off by more than 5%, use ``prtbad=5``. `prtbad`
        takes precedence over `prtbadh` and `prtbadl`.
    prtbadh : scalar or None; must be named; optional
        Only print rows where ``%diff > prtbadh``. Handy for showing
        just the exceedances. `prtbadh` takes precedence over
        `prtbadl`.
    prtbadl : scalar or None; must be named; optional
        Only print rows where ``%diff < prtbadl``. Handy for showing
        where reference rows are higher.
    flagbad : scalar or None; must be named; optional
        Flag % diffs where ``abs(%diff) > flagbad``. Works similar to
        `prtbad`. The flag is an asterisk (*).
    flagbadh : scalar or None; must be named; optional
        Flag % diffs where ``%diff > flagbadh``. Works similar to
        `prtbadh`. Handy for flagging exceedances. `flagbadh` takes
        precedence over `flagbadl`.
    flagbadl : scalar or None; must be named; optional
        Flag % diffs where ``%diff < flagbadl``. Works similar to
        `prtbadl`.
    dohistogram : bool; must be named; optional
        If True, plot the histograms. Plots will be written to
        "`filename`.histogram.png".
    histogram_inc : scalar; must be named; optional
        The histogram increment; defaults to 1.0 (for 1%).
    domagpct : bool; must be named; optional
        If True, plot the percent differences versus magnitude via
        :func:`magpct`. Plots will be written to
        "`filename`.magpct.png". Filtering for the "magpct" plot is
        controlled by the ``magpct_options['filterval']`` and
        ``magpct_options['symlogy']`` options. By default, all percent
        differences are shown, but the larger values (according to the
        `filterval` filter) are emphasized by using a mixed linear/log
        y-axis. The percent differences for the `ignorepv` rows are
        not plotted.
    magpct_options : None or dict; must be named; optional
        If None, it is internally reset to::

            magpct_options = {'filterval': 'filterval'}

        Use this parameter to provide any options to :func:`magpct`
        but note that the `filterval` option for :func:`magpct` is
        treated specially. Here, in addition to any of the values that
        :func:`magpct` accepts, it can also be set to the string
        "filterval" as in the default case shown above. In that case,
        ``magpct_options['filterval']`` gets internally reset to the
        initial value of `filterval` (described above).

        .. note::
            The call to :func:`magpct` is *after* applying `ignorepv`
            and doing any data aligning by labels.

        .. note::
           The two filter value options (`filterval` and
           ``magpct_options['filterval']``) have different defaults:
           None and 'filterval`, respectively. They also differ on how
           the ``None`` setting is used: for `filterval`, None is
           replaced by 1.e-6 while for `magpct_filterval`, None means
           that the "magpct" plot will not have any filters applied at
           all.

    doabsmax : bool; must be named; optional
        If True, compare only absolute maximums.
    shortabsmax : bool; must be named; optional
        If True, set ``doabsmax=True`` and do not print the max1 and
        min1 columns.
    roundvals : integer; must be named; optional
        Round max & min numbers at specified decimal. If negative, no
        rounding.
    rowhdr : string; must be named; optional
        Header for row number column
    deschdr : string; must be named; optional
        Header for description column
    maxhdr : string; must be named; optional
        Header for the column 1 data
    minhdr : string; must be named; optional
        Header for the column 2 data
    absmhdr : string; must be named; optional
        Header for abs-max column
    perpage : integer; must be named; optional
        The number of lines to write perpage. If < 1, there is no
        limit (one page).
    tight_layout_args : dict or None; must be named; optional
        Arguments for :func:`matplotlib.pyplot.tight_layout`. If None,
        defaults to ``{'pad': 3.0}``.
    show_figures : bool; must be named; optional
        If True, plot figures will be displayed on the screen for
        interactive viewing. Warning: there may be many figures.
    align_by_label : bool; must be named; optional
        If True, use labels to align the two sets of data for
        comparison. See note above under the `mxmn2` option.

    Returns
    -------
    pdiff_info : dict
        Dictionary with 'amx' (abs-max), 'mx' (max), and 'mn' keys:

        .. code-block:: none

            <class 'dict'>[n=3]
                'amx': <class 'dict'>[n=5]
                    'hsto' : float64 ndarray 33 elems: (11, 3)
                    'mag'  : [n=2]: (float64 ndarray: (100,), ...
                    'pct'  : float64 ndarray 100 elems: (100,)
                    'prtpv': bool ndarray 100 elems: (100,)
                    'spct' : [n=100]: ['  -2.46', '  -1.50', ...
                'mn' : <class 'dict'>[n=5]
                    'hsto' : float64 ndarray 33 elems: (11, 3)
                    'mag'  : [n=2]: (float64 ndarray: (100,), ...
                    'pct'  : float64 ndarray 100 elems: (100,)
                    'prtpv': bool ndarray 100 elems: (100,)
                    'spct' : [n=100]: ['   1.55', '   1.53', ...
                'mx' : <class 'dict'>[n=5]
                    'hsto' : float64 ndarray 27 elems: (9, 3)
                    'mag'  : [n=2]: (float64 ndarray: (100,), ...
                    'pct'  : float64 ndarray 100 elems: (100,)
                    'prtpv': bool ndarray 100 elems: (100,)
                    'spct' : [n=100]: ['  -2.46', '  -1.50', ...

        Where:

        .. code-block:: none

            'hsto'  : output of :func:`histogram`: [center, count, %]
            'mag'   : inputs to :func:`magpct`
            'pct'   : percent differences
            'prtpv' : rows to print partition vector
            'spct'  : string version of 'pct'

    Examples
    --------
    >>> import numpy as np
    >>> from pyyeti import cla
    >>> ext1 = [[120.0, -8.0],
    ...         [8.0, -120.0]]
    >>> ext2 = [[115.0, -5.0],
    ...         [10.0, -125.0]]

    Run :func:`rptpct1` multiple times to get a more complete picture
    of all the output (the table is very wide). Also, the plots will
    be turned off for this example.

    First, the header:

    >>> opts = {'domagpct': False, 'dohistogram': False}
    >>> dct = cla.rptpct1(ext1, ext2, 1, **opts)  # doctest: +ELLIPSIS
    PERCENT DIFFERENCE REPORT
    <BLANKLINE>
    Description: No description provided
    Uncertainty: Not specified
    Units:       Not specified
    Filter:      1e-06
    Notes:       % Diff = +/- abs(Self-Reference)/max(abs(Reference...
                 Sign set such that positive % differences indicate...
    Date:        ...
    ...

    Then, the max/min/absmax percent difference table in 3 calls:

    >>> dct = cla.rptpct1(ext1, ext2, 1, **opts)  # doctest: +ELLIPSIS
    PERCENT DIFFERENCE REPORT
    ...
                                 Self        Reference             ...
      Row    Description       Maximum        Maximum      % Diff  ...
    -------  -----------    -------------  -------------  -------  ...
          1  Row      1         120.00000      115.00000     4.35  ...
          2  Row      2           8.00000       10.00000    -1.60  ...
    ...
    >>> dct = cla.rptpct1(ext1, ext2, 1, **opts)  # doctest: +ELLIPSIS
    PERCENT DIFFERENCE REPORT
    ...
                         ...     Self        Reference             ...
      Row    Description ...   Minimum        Minimum      % Diff  ...
    -------  ----------- ...-------------  -------------  -------  ...
          1  Row      1  ...     -8.00000       -5.00000     2.61  ...
          2  Row      2  ...   -120.00000     -125.00000    -4.00  ...
    ...
    >>> dct = cla.rptpct1(ext1, ext2, 1, **opts)  # doctest: +ELLIPSIS
    PERCENT DIFFERENCE REPORT
    ...
                         ...     Self        Reference
      Row    Description ...   Abs-Max        Abs-Max      % Diff
    -------  ----------- ...-------------  -------------  -------
          1  Row      1  ...    120.00000      115.00000     4.35
          2  Row      2  ...    120.00000      125.00000    -4.00
    ...

    Finally, the histogram summaries:

    >>> dct = cla.rptpct1(ext1, ext2, 1, **opts)  # doctest: +ELLIPSIS
    PERCENT DIFFERENCE REPORT
    ...
        No description provided - Maximum Comparison Histogram
    <BLANKLINE>
          % Diff      Count    Percent
         --------   --------   -------
            -2.00          1     50.00
             4.00          1     50.00
    <BLANKLINE>
        0.0% of values are within 1%
        50.0% of values are within 2%
        100.0% of values are within 5%
    <BLANKLINE>
        % Diff Statistics: [Min, Max, Mean, StdDev] = [-1.60, 4.35,...
    <BLANKLINE>
    <BLANKLINE>
        No description provided - Minimum Comparison Histogram
    <BLANKLINE>
          % Diff      Count    Percent
         --------   --------   -------
            -4.00          1     50.00
             3.00          1     50.00
    <BLANKLINE>
        0.0% of values are within 1%
        100.0% of values are within 5%
    <BLANKLINE>
        % Diff Statistics: [Min, Max, Mean, StdDev] = [-4.00, 2.61,...
    <BLANKLINE>
    <BLANKLINE>
        No description provided - Abs-Max Comparison Histogram
    <BLANKLINE>
          % Diff      Count    Percent
         --------   --------   -------
            -4.00          1     50.00
             4.00          1     50.00
    <BLANKLINE>
        0.0% of values are within 1%
        100.0% of values are within 5%
    <BLANKLINE>
        % Diff Statistics: [Min, Max, Mean, StdDev] = [-4.00, 4.35,...
    """
    if tight_layout_args is None:
        tight_layout_args = {"pad": 3.0}

    if magpct_options is None:
        magpct_options = {"filterval": "filterval"}
    else:
        magpct_options = magpct_options.copy()

    # magpct_options['filterval'] get special treatment:
    magpct_filterval = magpct_options["filterval"]
    del magpct_options["filterval"]

    if isinstance(magpct_filterval, str):
        if magpct_filterval != "filterval":
            raise ValueError(
                "``magpct_options['filterval']`` is an invalid "
                f"string: {magpct_filterval!r} (can only "
                "be 'filterval' if a string)"
            )
        # copy the initial `filterval` setting:
        magpct_filterval = filterval

    infovars = (
        "desc",
        "filterval",
        "magpct_filterval",
        "labels",
        "units",
        "ignorepv",
        "uf_reds",
    )
    dct = locals()
    infodct = {n: dct[n] for n in infovars}
    del dct

    # check mxmn1:
    if isinstance(mxmn1, SimpleNamespace):
        sns = mxmn1.drminfo
        for key, value in infodct.items():
            if value is None:
                infodct[key] = getattr(sns, key, None)
        del sns
        mxmn1 = mxmn1.ext
    else:
        mxmn1 = np.atleast_2d(mxmn1)
    row_number = np.arange(1, mxmn1.shape[0] + 1)

    # check mxmn2:
    if isinstance(mxmn2, SimpleNamespace) and getattr(mxmn2, "drminfo", None):
        labels2 = mxmn2.drminfo.labels
        mxmn2 = mxmn2.ext
        if align_by_label:
            # use labels and labels2 to align data; this is in case
            # the two sets of results recover some of the same items,
            # but not all
            mxmn1, mxmn2, row_number = _align_mxmn(
                mxmn1, mxmn2, labels2, row_number, infodct
            )
    else:
        mxmn2 = np.atleast_2d(mxmn2)

    desc = infodct["desc"]
    if desc is None:
        desc = "No description provided"

    R = mxmn1.shape[0]
    if R != mxmn2.shape[0]:
        raise ValueError(
            f"`mxmn1` and `mxmn2` have a different number of rows: "
            f"{R} vs {mxmn2.shape[0]} for category with `desc` = {desc}"
        )

    filterval = infodct["filterval"]
    magpct_filterval = infodct["magpct_filterval"]
    labels = infodct["labels"]
    units = infodct["units"]
    ignorepv = infodct["ignorepv"]
    uf_reds = infodct["uf_reds"]
    del infodct

    if filterval is None:
        filterval = 1.0e-6
    filterval = _proc_filterval(filterval, R, "filterval")
    magpct_filterval = _proc_filterval(
        magpct_filterval, R, "magpct_options['filterval']"
    )

    if labels is None:
        labels = [f"Row {i + 1:6d}" for i in range(R)]
    elif len(labels) != R:
        raise ValueError(
            "length of `labels` does not match number"
            f" of rows in `mxmn1`: {len(labels)} vs {R} for "
            f"category with `desc` = {desc}"
        )
    if units is None:
        units = "Not specified"
    if numform is None:
        numform = _get_numform(mxmn1)

    pdhdr = "% Diff"
    nastring = "n/a "
    comppv = np.ones(R, bool)
    if ignorepv is not None:
        comppv[ignorepv] = False

    # for row labels:
    w = max(11, len(max(labels, key=len)))
    frm = f"{{:{w}}}"

    # start preparing for writer.formheader:
    print_info = SimpleNamespace(
        headers1=["", ""],
        headers2=[rowhdr, deschdr],
        formats=["{:7d}", frm],
        printargs=[row_number, labels],
        widths=[7, w],
        seps=[0, 2],
        justs=["c", "l"],
    )

    if shortabsmax:
        doabsmax = True
    if doabsmax:
        use_range = False
    if roundvals > -1:
        mxmn1 = np.round(mxmn1, roundvals)
        mxmn2 = np.round(mxmn2, roundvals)

    prtbads = (prtbad, prtbadh, prtbadl)
    flagbads = (flagbad, flagbadh, flagbadl)

    # compute percent differences
    pctinfo = {}
    kwargs = dict(
        names=names,
        mxmn1=mxmn1,
        comppv=comppv,
        histogram_inc=histogram_inc,
        numform=numform,
        prtbads=prtbads,
        flagbads=flagbads,
        maxhdr=maxhdr,
        minhdr=minhdr,
        absmhdr=absmhdr,
        pdhdr=pdhdr,
        nastring=nastring,
        doabsmax=doabsmax,
        shortabsmax=shortabsmax,
        print_info=print_info,
    )
    with warnings.catch_warnings():
        warnings.filterwarnings("ignore", r"All-NaN (slice|axis) encountered")
        mx1 = np.nanmax(abs(mxmn1), axis=1)
        mx2 = np.nanmax(abs(mxmn2), axis=1)
    if not doabsmax:
        max1, min1 = mxmn1[:, 0], mxmn1[:, 1]
        max2, min2 = mxmn2[:, 0], mxmn2[:, 1]
        mxmn_b = mxmn2 if use_range else None
        prtpv = np.zeros(R, bool)
        for i in zip(
            ("mx", "mn", "amx"),
            (max1, min1, mx1),
            (max2, min2, mx2),
            (True, False, True),
            (maxhdr, minhdr, absmhdr),
        ):
            lbl, ext1, ext2, ismax, valhdr = i
            pctinfo[lbl] = _proc_pct(
                ext1,
                ext2,
                filterval,
                magpct_filterval,
                mxmn_b=mxmn_b,
                ismax=ismax,
                valhdr=valhdr,
                **kwargs,
            )
            prtpv |= pctinfo[lbl]["prtpv"]
        prtpv &= comppv
    else:
        pctinfo["amx"] = _proc_pct(
            mx1,
            mx2,
            filterval,
            magpct_filterval,
            mxmn_b=None,
            ismax=True,
            valhdr=absmhdr,
            **kwargs,
        )
        prtpv = pctinfo["amx"]["prtpv"]
    hu, frm = writer.formheader(
        [print_info.headers1, print_info.headers2],
        print_info.widths,
        print_info.formats,
        sep=print_info.seps,
        just=print_info.justs,
    )

    # format page header:
    misc = _get_filtline(filterval) + _get_noteline(use_range, names, prtbads, flagbads)
    hdrs = _get_rpt_headers(desc=desc, uf_reds=uf_reds, units=units, misc=misc)
    header = title + "\n\n" + hdrs + "\n"

    imode = plt.isinteractive()
    plt.interactive(show_figures)
    try:
        if domagpct:
            _plot_magpct(
                pctinfo,
                names,
                desc,
                doabsmax,
                filename,
                magpct_options,
                use_range,
                maxhdr,
                minhdr,
                absmhdr,
                show_figures,
                tight_layout_args,
            )
        if dohistogram:
            _plot_histogram(
                pctinfo,
                names,
                desc,
                doabsmax,
                filename,
                histogram_inc,
                maxhdr,
                minhdr,
                absmhdr,
                show_figures,
                tight_layout_args,
            )
    finally:
        plt.interactive(imode)

    # write results
    @ytools.write_text_file
    def _wtcmp(f, header, hu, frm, printargs, perpage, prtpv, pctinfo, desc):
        prtpv = prtpv.nonzero()[0]
        if perpage < 1:
            # one additional in case size is zero
            perpage = prtpv.size + 1
        pages = (prtpv.size + perpage - 1) // perpage
        if prtpv.size < len(printargs[0]):
            for i, item in enumerate(printargs):
                printargs[i] = [item[j] for j in prtpv]
        tabhead = header + hu
        pager = "\n"  # + chr(12)
        for p in range(pages):
            if p > 0:
                f.write(pager)
            f.write(tabhead)
            b = p * perpage
            e = b + perpage
            writer.vecwrite(f, frm, *printargs, so=slice(b, e))
        f.write(pager)
        for lbl, hdr in zip(("mx", "mn", "amx"), (maxhdr, minhdr, absmhdr)):
            if lbl in pctinfo:
                f.write(_get_histogram_str(desc, hdr, pctinfo[lbl]))

    _wtcmp(
        filename, header, hu, frm, print_info.printargs, perpage, prtpv, pctinfo, desc
    )
    return pctinfo



# -*- coding: utf-8 -*-
"""
Python tools for reading/writing Nastran .op4 files. Can read and
write all formats (as far as I know) with the restrictions that the
output files created by this class are always double precision, and
all matrices are read in as double precision. The binary files can be
in big or little endian format.

Notes on sparse matrices:

  1. By default, matrices read from .op4 files will be regular
     :class:`numpy.ndarray` matrices. However, :mod:`scipy.sparse`
     matrices can be created instead. See the `sparse` option in
     :func:`read` and :func:`load`.

  2. By default, matrices written to .op4 files will follow the Python
     type: :class:`numpy.ndarray` matrices will be written in dense
     format and :mod:`scipy.sparse` matrices will be written in
     "bigmat" sparse format. This can be overridden by specifying the
     `sparse` option in :func:`write`.

.. note::
    Some features of this module are demonstrated in the pyYeti
    :ref:`tutorial`: :doc:`/tutorials/op4`. There is also a link to
    the source Jupyter notebook at the top of the tutorial.
"""

import itertools as it
import struct
import sys
import warnings
import collections
import numpy as np
import scipy.sparse as sp
from pyyeti import guitools


class OP4:
    """
    Class for reading/writing Nastran output4 (.op4) files.

    See demo below and refer to the help on these functions for more
    information: :func:`write` (or :func:`save`), :func:`load` (or the 
    lower level :func:`dctload`, :func:`listload`), and
    :func:`dir`. `save` is an alias for `write`.

    Examples
    --------
    Instantiate the class and create matrices for demo:

    >>> from pyyeti.nastran import op4
    >>> o4 = op4.OP4()
    >>> import numpy as np
    >>> r = np.random.randn(3, 5)
    >>> c = 1j*np.random.randn(3, 5) + r

    Write binary op4 file, with 'r' first:

    >>> o4.write('testbin.op4', ['r', 'c'], [r, c])

    Write ascii op4 file without caring about order:

    >>> o4.write('testascii.op4', dict(r=r, c=c), binary=False)

    To read an op4 file into a dictionary (indexed by the name in
    lower case):

    >>> dct = o4.load('testbin.op4', into='dct')

    Note: to preserve order, the dictionary returned by "load" or
    "read" is actually an OrderedDict from the standard Python
    "collections" module (:class:`collections.OrderedDict`).

    To read into a list:

    >>> names, mats, forms, mtypes = o4.load('testascii.op4',
    ...                                      into='list')

    Check some results:

    >>> print(np.all(r == dct['r'][0]))
    True

    >>> if names[0] == 'c':
    ...     print(np.all(c == mats[0]))
    ... else:
    ...     print(np.all(c == mats[1]))
    True

    To print a 'directory' of an op4 file:

    >>> d = o4.dir('testbin.op4')
    r       ,      3 x 5     , form=2, mtype=2
    c       ,      3 x 5     , form=2, mtype=4

    Clean up:

    >>> import os
    >>> os.remove('testbin.op4')
    >>> os.remove('testascii.op4')
    """

    def __init__(self):
        self._fileh = None
        string = "%.1E" % 1.2
        self._expdigits = len(string) - (string.find("E") + 2)
        self._rows4bigmat = 65536
        # Tunable value ... if number of values exceeds this, read
        # with numpy.fromfile instead of struct.unpack.
        self._rowsCutoff = 3000
        self.save = self.write

    def __del__(self):
        if self._fileh:
            self._fileh.close()
            self._fileh = None

    def _op4close(self):
        if self._fileh:
            self._fileh.close()
            self._fileh = None

    def _op4open_read(self, filename):
        """
        Open binary or ascii op4 file for reading.

        Sets these class variables:

        _fileh : file handle
            Value returned by open(). File is opened in 'r' mode if
            ascii, 'rb' mode if binary.
        _ascii : bool
            True if file is ascii.
        _dformat : bool
            True if an ascii file uses 'D' instead of 'E' (eg, 1.4D3
            instead of 1.4E3)
        _bit64 : True or False
            True if 'key' integers are 64-bit in binary op4 files.
        _endian : string
            Will be '' if no byte-swapping is required; otherwise,
            either '>' or '<' for big-endian and little-endian,
            respectively. Only used for binary files.
        _Str_i4 : struct.Struct object
            Precompiled for reading 4 byte integers
        _Str_i : struct.Struct object
            Precompiled for reading 4 or 8 byte integers
        _bytes_i : integer
            Either 4 or 8, to go with Str_i.
        _str_sr : string
            Either self._endian + '%df' or self._endian + '%dd',
            depending on self._bit64; for reading single precision
            reals.
        _bytes_sr : integer
            Number of bytes in single real.
        _str_dr : string
            self._endian + '%dd', for reading double precision reals.
        _wordsperdouble : integer
            Either 1 or 2; 2 if self._bit64 is False.
        """
        self._fileh = open(filename, "rb")
        bytes = self._fileh.read(16)
        self._endian = ""
        self._dformat = False
        self._matcount = 0

        # Assuming binary, check for a zero byte in the 'type' field;
        # will have one at front or back if binary:
        if bytes[12] == 0 or bytes[15] == 0:
            self._ascii = False
            if sys.byteorder == "little":
                if bytes[12] == 0:
                    self._endian = ">"
            else:
                if bytes[12] != 0:
                    self._endian = "<"
            self._Str_i4 = struct.Struct(self._endian + "i")
            reclen = self._Str_i4.unpack(bytes[:4])[0]
            if reclen == 48:
                self._bit64 = True
                self._Str_i = struct.Struct(self._endian + "q")
                self._bytes_i = 8
                self._Str_ii = struct.Struct(self._endian + "qq")
                self._bytes_ii = 16
                self._Str_iii = struct.Struct(self._endian + "3q")
                self._bytes_iii = 24
                self._Str_iiii = struct.Struct(self._endian + "4q")
                self._bytes_iiii = 32
                self._str_sr = self._endian + "%dd"
                self._str_sr_fromfile = np.dtype(self._endian + "f8")
                self._bytes_sr = 8
                self._wordsperdouble = 1
            else:
                self._bit64 = False
                self._Str_i = self._Str_i4
                self._bytes_i = 4
                self._Str_ii = struct.Struct(self._endian + "ii")
                self._bytes_ii = 8
                self._Str_iii = struct.Struct(self._endian + "3i")
                self._bytes_iii = 12
                self._Str_iiii = struct.Struct(self._endian + "4i")
                self._bytes_iiii = 16
                self._str_sr = self._endian + "%df"
                self._str_sr_fromfile = np.dtype(self._endian + "f4")
                self._bytes_sr = 4
                self._wordsperdouble = 2
            self._str_dr = self._endian + "%dd"
            self._str_dr_fromfile = np.dtype(self._endian + "f8")
            self._fileh.seek(0)
        else:
            self._ascii = True
            self._fileh.readline()
            self._fileh.readline()
            line = self._fileh.readline().decode()
            # sparse formats have integer header line:
            if line.find(".") == -1:
                line = self._fileh.readline().decode()
            if line.find("D") > -1:
                self._dformat = True
            self._fileh.close()
            self._fileh = open(filename, "r")

    def _skipop4_ascii(self, perline, rows, cols, mtype):
        """
        Skip an op4 matrix - ascii.

        Parameters
        ----------
        perline : integer
            Number of elements per line in the file.
        rows : integer
            Number of rows in matrix.
        cols : integer
            Number of columns in matrix.
        mtype : integer
            Nastran matrix type.

        Returns
        -------
        None

        On entry, file is positioned after the title line, but before
        the first column is printed. On exit, the file is positioned
        so the next readline will get the next title line.
        """
        # read until next matrix:
        bigmat = rows < 0 or rows >= self._rows4bigmat
        if mtype & 1:
            wper = 1
        else:
            wper = 2
        line = self._fileh.readline()
        c = int(line[:8]) - 1
        r = int(line[8:16])
        if r > 0:
            while c < cols:
                elems = int(line[16:24])
                nlines = (elems + perline - 1) // perline
                for _ in it.repeat(None, nlines):
                    self._fileh.readline()
                line = self._fileh.readline()
                c = int(line[:8]) - 1
        elif bigmat:
            while c < cols:
                elems = int(line[16:24])
                while elems > 0:
                    line = self._fileh.readline()
                    L = int(line[:8]) - 1  # L
                    elems -= L + 2
                    L //= wper
                    # read column as a long string
                    nlines = (L + perline - 1) // perline
                    for _ in it.repeat(None, nlines):
                        self._fileh.readline()
                line = self._fileh.readline()
                c = int(line[:8]) - 1
        else:
            while c < cols:
                elems = int(line[16:24])
                while elems > 0:
                    line = self._fileh.readline()
                    IS = int(line[:8])
                    L = (IS >> 16) - 1  # L
                    elems -= L + 1
                    L //= wper
                    # read column as a long string
                    nlines = (L + perline - 1) // perline
                    for _ in it.repeat(None, nlines):
                        self._fileh.readline()
                line = self._fileh.readline()
                c = int(line[:8]) - 1
        self._fileh.readline()

    def _check_name(self, name):
        """
        Check name read from op4 file: strip all blanks/nulls; if name
        is not a valid Python identifier, it is set to
        f"m{self._matcount}". self._matcount starts at 0 and is
        incremented in this routine.

        Returns new name (usually the same as the input name).
        """
        name = name.strip(" \x00").replace(" ", "").replace("\x00", "").lower()
        if not name.isidentifier():
            oldname, name = name, f"m{self._matcount}"
            warnings.warn(
                f"Output4 file has matrix name: {oldname!r}. Changing to {name!r}.",
                RuntimeWarning,
            )
        self._matcount += 1
        return name

    def _get_ascii_block(self, L, perline, linelen):
        fh = self._fileh
        nlines = (L - 1) // perline + 1
        blocklist = [ln[:linelen] for ln in it.islice(fh, nlines)]
        s = "".join(blocklist)
        if self._dformat:
            s = s.replace("D", "E")
        return s

    @staticmethod
    def _init_dense_real(rows, cols):
        return np.zeros((rows, cols), dtype=float, order="F")

    @staticmethod
    def _init_dense_complex(rows, cols):
        return np.zeros((rows, cols), dtype=complex, order="F")

    @staticmethod
    def _put_ascii_values(X, r, c, s, L, numlen):
        a = 0
        for i in range(L):
            b = a + numlen
            X[r + i, c] = s[a:b]
            a = b

    @staticmethod
    def _put_ascii_values_c(X, r, c, s, L, numlen):
        a = 0
        for i in range(L // 2):
            b = a + numlen
            real = float(s[a:b])
            a = b
            b = a + numlen
            imag = float(s[a:b])
            a = b
            X[r + i, c] = real + 1j * imag

    @staticmethod
    def _dense_matrix(rows, cols, X):
        return X

    @staticmethod
    def _init_sparse(rows, cols):
        return [], [], []

    @staticmethod
    def _put_ascii_values_sparse(X, r, c, s, L, numlen):
        I, J, V = X
        a = 0
        for i in range(L):
            b = a + numlen
            I.append(r + i)
            J.append(c)
            V.append(float(s[a:b]))
            a = b

    @staticmethod
    def _put_ascii_values_sparse_c(X, r, c, s, L, numlen):
        I, J, V = X
        a = 0
        for i in range(L // 2):
            b = a + numlen
            real = float(s[a:b])
            a = b
            b = a + numlen
            imag = float(s[a:b])
            a = b
            I.append(r + i)
            J.append(c)
            V.append(real + 1j * imag)

    @staticmethod
    def _sparse_matrix(rows, cols, X):
        I, J, V = X
        return sp.coo_matrix((np.array(V), (I, J)), shape=(rows, cols))

    def _rd_dense_ascii(
        self, wper, r, c, rows, cols, line, numlen, perline, linelen, funcs
    ):
        init, put, retrn = funcs
        X = init(rows, cols)
        while c < cols:
            elems = int(line[16:24])
            r -= 1
            # read column as a long string
            s = self._get_ascii_block(elems, perline, linelen)
            put(X, r, c, s, elems, numlen)
            line = self._fileh.readline()
            c = int(line[:8]) - 1
            r = int(line[8:16])
        return retrn(rows, cols, X)

    def _rd_bigmat_ascii(
        self, wper, r, c, rows, cols, line, numlen, perline, linelen, funcs
    ):
        init, put, retrn = funcs
        X = init(rows, cols)
        while c < cols:
            elems = int(line[16:24])
            while elems > 0:
                line = self._fileh.readline()
                L = int(line[:8]) - 1  # L
                r = int(line[8:16]) - 1  # irow-1
                elems -= L + 2
                L //= wper
                s = self._get_ascii_block(L, perline, linelen)
                put(X, r, c, s, L, numlen)
            line = self._fileh.readline()
            c = int(line[:8]) - 1
            # r = int(line[8:16])
        return retrn(rows, cols, X)

    def _rd_nonbigmat_ascii(
        self, wper, r, c, rows, cols, line, numlen, perline, linelen, funcs
    ):
        init, put, retrn = funcs
        X = init(rows, cols)
        while c < cols:
            elems = int(line[16:24])
            while elems > 0:
                line = self._fileh.readline()
                IS = int(line)  # [:8])
                L = (IS >> 16) - 1  # L
                r = IS - ((L + 1) << 16) - 1  # irow-1
                elems -= L + 1
                L //= wper
                s = self._get_ascii_block(L, perline, linelen)
                put(X, r, c, s, L, numlen)
            line = self._fileh.readline()
            c = int(line[:8]) - 1
            # r = int(line[8:16])
        return retrn(rows, cols, X)

    def _get_funcs(self, a_or_b, rows, r, mtype, sparse):
        if a_or_b == "ascii":
            rd_dense = self._rd_dense_ascii
            rd_bigmat = self._rd_bigmat_ascii
            rd_nonbigmat = self._rd_nonbigmat_ascii
            put_values = OP4._put_ascii_values
            put_values_c = OP4._put_ascii_values_c
            put_values_sparse = OP4._put_ascii_values_sparse
            put_values_sparse_c = OP4._put_ascii_values_sparse_c
        else:
            rd_dense = self._rd_dense_binary
            rd_bigmat = self._rd_bigmat_binary
            rd_nonbigmat = self._rd_nonbigmat_binary
            put_values = OP4._put_binary_values
            put_values_c = OP4._put_binary_values_c
            put_values_sparse = OP4._put_binary_values_sparse
            put_values_sparse_c = OP4._put_binary_values_sparse_c

        # if self._rows4bigmat > rows > 0 and r > 0:
        if rows > 0 and r > 0:
            # dense format
            rdfunc = rd_dense
            if sparse is None:
                sparse = False
        else:
            # either bigmat or nonbigmat sparse format:
            if sparse is None:
                sparse = True
            if rows < 0 or rows >= self._rows4bigmat:
                # bigmat sparse format
                rdfunc = rd_bigmat
            else:
                # nonbigmat sparse format
                rdfunc = rd_nonbigmat

        if not sparse:
            if mtype < 3:
                funcs = (OP4._init_dense_real, put_values, OP4._dense_matrix)
            else:
                # uses "real" put function because matrix is read as
                # real and converted at the end to complex via a view
                funcs = (OP4._init_dense_complex, put_values_c, OP4._dense_matrix)
        else:
            if mtype < 3:
                put = put_values_sparse
            else:
                put = put_values_sparse_c
            funcs = (OP4._init_sparse, put, OP4._sparse_matrix)
        return rdfunc, funcs

    @staticmethod
    def _get_sparsefunc(sparse):
        try:
            sparse, sparsefunc = sparse
        except TypeError:
            sparsefunc = None
        return sparse, sparsefunc

    def _loadop4_ascii(self, patternlist=None, listonly=False, sparse=False):
        """
        Reads next matching matrix or returns information on the next
        matrix in the ascii op4 file.

        Parameters
        ----------
        patternlist : list
            List of string patterns; each matrix name is matched
            against this list:  if it matches any of the patterns, it
            is read in.
        listonly : bool
            True if only reading name.
        sparse : bool or None or two-tuple_like; optional
            Specifies whether output matrices will be regular numpy
            arrays or sparse arrays. If not tuple_like:

            ========   ===============================================
            `sparse`   Action
            ========   ===============================================
             None      Auto setting: each matrix will be sparse if and
                       only if it was written in a sparse format
             True      Matrices will be returned in sparse format
             False     Matrices will be returned in regular (dense)
                       numpy arrays
            ========   ===============================================

            If `sparse` is two-tuple_like, the first element is either
            None, True, or False (see table above) and the second
            element is a callable, as in: ``X = callable(X)``. A
            common usage of the callable would be to convert from
            "COO" sparse form (see :class:`scipy.sparse.coo_matrix`)
            to a more desirable form. For example, to ensure *all*
            matrices are returned in CSC form (see
            :class:`scipy.sparse.csc_matrix`) use::

                sparse=(True, scipy.sparse.coo_matrix.tocsc)

            The callable is ignored for non-sparse matrices.

        Returns
        -------
        tuple: (name, matrix, form, mtype)
            name : string
                Lower-case name of matrix.
            matrix : 2d ndarray
                The matrix.
            form : integer
                Nastran form of matrix.
            mtype : integer
                Nastran matrix type.

        Notes
        -----
        - All outputs will be None if reached EOF.
        - The `matrix` output will be [rows, cols] of the matrix if
          the matrix is skipped.
        - The default form for sparse matrices is the "COO" sparse
          form (see :class:`scipy.sparse.coo_matrix`). To override,
          provide a callable in the `sparse` option (see above).
        """
        while 1:
            line = self._fileh.readline()
            line = line.rstrip()
            if line == "":
                return None, None, None, None
            cols = int(line[:8])
            rows = int(line[8:16])
            form = int(line[16:24])
            mtype = int(line[24:32])
            length = len(line)
            name = self._check_name(line[32:40])
            perline = 5
            numlen = 16
            if length > 44:
                # 1P,3E24.16  <-- starts at position 40
                # 3e24.16
                numformat = line[40:].strip().upper()
                if numformat.startswith("1P,"):
                    numformat = numformat[3:]
                p = numformat.replace("D", "E").find("E")
                if p > 0:
                    perline = int(numformat[:p])
                    numlen = int(numformat[p + 1 :].split(".")[0])
            if patternlist and name not in patternlist:
                skip = 1
            else:
                skip = 0
            if listonly or skip:
                self._skipop4_ascii(perline, rows, cols, mtype)
                if listonly:
                    return name, (abs(rows), cols), form, mtype
            else:
                break

        wper = 1 if mtype & 1 else 2
        line = self._fileh.readline()
        linelen = perline * numlen
        c = int(line[:8]) - 1
        r = int(line[8:16])
        sparse, sparsefunc = OP4._get_sparsefunc(sparse)

        rdfunc, funcs = self._get_funcs("ascii", rows, r, mtype, sparse)
        X = rdfunc(wper, r, c, abs(rows), cols, line, numlen, perline, linelen, funcs)

        if sparsefunc and sp.issparse(X):
            X = sparsefunc(X)
        self._fileh.readline()
        return name, X, form, mtype

    def _skipop4_binary(self, cols):
        """
        Skip a binary op4 matrix.

        Parameters
        ----------
        cols : integer
            Number of columns in matrix.
        """
        # Scan matrix by column
        icol = 1
        bi = self._bytes_i
        delta = 4 - bi
        while icol <= cols:
            # Read record length at start of record:
            reclen = self._Str_i4.unpack(self._fileh.read(4))[0]
            # Read column header
            icol = self._Str_i.unpack(self._fileh.read(bi))[0]
            self._fileh.seek(reclen + delta, 1)

    def _get_cutoff_etc(self):
        return (
            self._rowsCutoff,
            self._Str_iii.unpack,
            self._bytes_iii,
            self._Str_i4.unpack,
        )

    def _get_s2(self):
        return self._Str_ii.unpack, self._bytes_ii

    def _get_s1(self):
        return self._Str_i.unpack, self._bytes_i

    @staticmethod
    def _put_binary_values(X, r, c, Y):
        X[r : r + len(Y), c] = Y

    @staticmethod
    def _put_binary_values_c(X, r, c, Y):
        if not isinstance(Y, np.ndarray):
            Y = np.array(Y)
        Y = Y.astype("float", copy=False)
        Y.dtype = complex
        X[r : r + len(Y), c] = Y

    @staticmethod
    def _put_binary_values_sparse(X, r, c, Y):
        I, J, V = X
        for i, v in enumerate(Y):
            I.append(r + i)
            J.append(c)
            V.append(v)

    @staticmethod
    def _put_binary_values_sparse_c(X, r, c, Y):
        I, J, V = X
        for i, j in enumerate(range(0, len(Y), 2)):
            I.append(r + i)
            J.append(c)
            V.append(Y[j] + 1j * Y[j + 1])

    def _rd_dense_binary(
        self,
        fp,
        wper,
        r,
        c,
        rows,
        cols,
        nwords,
        reclen,
        bytesreal,
        numform,
        numform2,
        funcs,
    ):
        init, put, retrn = funcs
        X = init(rows, cols)
        cutoff, s3, b3, s4 = self._get_cutoff_etc()
        while c < cols:
            r -= 1
            nwords //= wper
            if nwords < cutoff:
                Y = struct.unpack(numform % nwords, fp.read(bytesreal * nwords))
            else:
                Y = np.fromfile(fp, numform2, nwords)
            put(X, r, c, Y)
            fp.read(4)
            reclen = s4(fp.read(4))[0]
            c, r, nwords = s3(fp.read(b3))
            c -= 1
        return retrn(rows, cols, X), reclen

    def _rd_bigmat_binary(
        self,
        fp,
        wper,
        r,
        c,
        rows,
        cols,
        nwords,
        reclen,
        bytesreal,
        numform,
        numform2,
        funcs,
    ):
        init, put, retrn = funcs
        X = init(rows, cols)
        cutoff, s3, b3, s4 = self._get_cutoff_etc()
        s2, b2 = self._get_s2()
        while c < cols:
            # bigmat sparse format
            # Read column data, one string of numbers at a time
            # (strings of zeros are skipped)
            while nwords > 0:
                L, r = s2(fp.read(b2))
                nwords -= L + 1
                L = (L - 1) // wper
                r -= 1
                if L < cutoff:
                    Y = struct.unpack(numform % L, fp.read(bytesreal * L))
                else:
                    Y = np.fromfile(fp, numform2, L)
                put(X, r, c, Y)
            fp.read(4)
            reclen = s4(fp.read(4))[0]
            c, r, nwords = s3(fp.read(b3))
            c -= 1
        return retrn(rows, cols, X), reclen

    def _rd_nonbigmat_binary(
        self,
        fp,
        wper,
        r,
        c,
        rows,
        cols,
        nwords,
        reclen,
        bytesreal,
        numform,
        numform2,
        funcs,
    ):
        init, put, retrn = funcs
        X = init(rows, cols)
        cutoff, s3, b3, s4 = self._get_cutoff_etc()
        s1, b1 = self._get_s1()
        while c < cols:
            # non-bigmat sparse format
            # Read column data, one string of numbers at a time
            # (strings of zeros are skipped)
            while nwords > 0:
                IS = s1(fp.read(b1))[0]
                L = (IS >> 16) - 1  # L
                r = IS - ((L + 1) << 16) - 1  # irow-1
                nwords -= L + 1  # words left
                L //= wper
                if L < cutoff:
                    Y = struct.unpack(numform % L, fp.read(bytesreal * L))
                else:
                    Y = np.fromfile(fp, numform2, L)
                put(X, r, c, Y)
            fp.read(4)
            reclen = s4(fp.read(4))[0]
            c, r, nwords = s3(fp.read(b3))
            c -= 1
        return retrn(rows, cols, X), reclen

    def _loadop4_binary(self, patternlist=None, listonly=False, sparse=False):
        """
        Reads next matching matrix or returns information on the next
        matrix in the binary op4 file.

        Parameters
        ----------
        patternlist : list
            List of string patterns; each matrix name is matched
            against this list:  if it matches any of the patterns, it
            is read in.
        listonly : bool
            True if only reading name.
        sparse : bool or None or two-tuple_like; optional
            Specifies whether output matrices will be regular numpy
            arrays or sparse arrays. If not two-tuple_like:

            ========   ===============================================
            `sparse`   Action
            ========   ===============================================
             None      Auto setting: each matrix will be sparse if and
                       only if it was written in a sparse format
             True      Matrices will be returned in sparse format
             False     Matrices will be returned in regular (dense)
                       numpy arrays
            ========   ===============================================

            If `sparse` is two-tuple_like, the first element is either
            None, True, or False (see table above) and the second
            element is a callable, as in: ``X = callable(X)``. A
            common usage of the callable would be to convert from
            "COO" sparse form (see :class:`scipy.sparse.coo_matrix`)
            to a more desirable form. For example, to ensure *all*
            matrices are returned in CSC form (see
            :class:`scipy.sparse.csc_matrix`) use::

                sparse=(True, scipy.sparse.coo_matrix.tocsc)

            The callable is ignored for non-sparse matrices.

        Returns
        -------
        tuple: (name, matrix, form, mtype)
            name : string
                Lower-case name of matrix.
            matrix : 2d ndarray
                The matrix.
            form : integer
                Nastran form of matrix.
            mtype : integer
                Nastran matrix type.

        Notes:
        - All outputs will be None if reached EOF.
        - The `matrix` output will be [rows, cols] of the matrix if
          the matrix is skipped.
        """
        fp = self._fileh
        while 1:
            if len(fp.read(4)) == 0:
                return None, None, None, None
            cols, rows, form, mtype = self._Str_iiii.unpack(fp.read(self._bytes_iiii))
            # Read ascii name of matrix:
            if self._bit64:
                name = fp.read(16).decode()
            else:
                name = fp.read(8).decode()
            name = self._check_name(name)
            fp.read(4)
            if patternlist and name not in patternlist:
                skip = 1
            else:
                skip = 0
            if listonly or skip:
                self._skipop4_binary(cols)
                if listonly:
                    return name, (abs(rows), cols), form, mtype
            else:
                break

        reclen = self._Str_i4.unpack(fp.read(4))[0]
        c, r, nwords = self._Str_iii.unpack(fp.read(self._bytes_iii))
        c -= 1
        sparse, sparsefunc = OP4._get_sparsefunc(sparse)

        rdfunc, funcs = self._get_funcs("binary", rows, r, mtype, sparse)
        if mtype & 1:
            numform = self._str_sr
            numform2 = self._str_sr_fromfile
            bytesreal = self._bytes_sr
            wper = 1
        else:
            numform = self._str_dr
            numform2 = self._str_dr_fromfile
            bytesreal = 8
            wper = self._wordsperdouble

        X, reclen = rdfunc(
            fp,
            wper,
            r,
            c,
            abs(rows),
            cols,
            nwords,
            reclen,
            bytesreal,
            numform,
            numform2,
            funcs,
        )

        if sparsefunc and sp.issparse(X):
            X = sparsefunc(X)

        # read final bytes of record and record marker
        nbytes = reclen - 3 * self._bytes_i + 4
        if len(fp.read(nbytes)) < nbytes:
            warnings.warn(
                f"Premature end-of-file after matrix {name!r}. Nastran "
                "will likely FATAL on this file.",
                RuntimeWarning,
            )
        return name, X, form, mtype

    @staticmethod
    def _sparse_col_stats(r):
        """
        Returns locations of non-zero values and length of each
        series.

        Parameters
        ----------
        r : ndarray
            1d ndarray containing the indices of the values in the
            data column

        Returns
        -------
        ind : ndarray
            m x 2 ndarray. m is number of non-zero sequences in v.
            First column contains the indices to the start of each
            sequence and the second column contains the length of
            the sequence.

        For example, if v is::

          v = [ 0.,  0.,  0.,  7.,  5.,  0.,  6.,  0.,  2.,  3.]

        which makes r be::

          r = [ 3, 4, 6, 8, 9 ]

        Then, ind will be::

          ind = [[3 2]
                 [6 1]
                 [8 2]]
        """
        dr = np.diff(r)
        starts = np.nonzero(dr != 1)[0] + 1
        nrows = len(starts) + 1
        ind = np.zeros((nrows, 2), int)
        ind[0, 0] = r[0]
        if nrows > 1:
            ind[1:, 0] = r[starts]
            ind[0, 1] = starts[0] - 1
            ind[1:-1, 1] = np.diff(starts) - 1
        ind[-1, 1] = len(dr) - len(starts) - sum(ind[:, 1])
        ind[:, 1] += 1
        return ind

    # @staticmethod
    # def _is_symmetric(m, tol=1e-12):
    #     """
    #     returns True if `m` is approx symmetric; sparse or not
    #     """
    #     return abs(m - m.transpose()).max() <= tol * abs(m).max()

    @staticmethod
    def _is_symmetric(m):
        """
        Returns True if matrix `m` is approx symmetric.

        Works for sparse and non-sparse matrices. Only called for
        square matrices, so that check is not done here.

        Note: if m is sparse, what gets input to the routine is a
        tuple: (m, r, c, v) where m is the original sparse matrix and
        r, c, v is the output from scipy.sparse.find(m).
        """
        if isinstance(m, tuple):
            r, c, v = m[1:]
            low = r > c  # values in lower triangle
            upp = c > r  # values in upper triangle

            if np.count_nonzero(low) != np.count_nonzero(upp):
                return False

            rl = r[low]
            cl = c[low]
            vl = v[low]
            ru = r[upp]
            cu = c[upp]
            vu = v[upp]

            sortl = np.lexsort((cl, rl))
            sortu = np.lexsort((ru, cu))
            return (
                np.all(cl[sortl] == ru[sortu])
                and np.all(rl[sortl] == cu[sortu])
                and np.allclose(vl[sortl], vu[sortu])
            )
        return np.allclose(m.transpose(), m)

    @staticmethod
    def _sparse_sort(matrix):
        # sparse matrix:
        m, r, c, v = matrix
        pv = np.lexsort((r, c))  # sort by column, then by row
        rs = r[pv]
        cs = c[pv]
        vs = v[pv]
        cols_with_data = sorted(set(cs))
        return rs, cs, vs, cols_with_data

    @staticmethod
    def _get_header_info(matrix, form=None):
        if isinstance(matrix, tuple):
            mat = matrix[0]
            vals = matrix[3]
        else:
            mat = vals = matrix
        rows, cols = mat.shape
        if form is None:
            if rows == cols:
                if OP4._is_symmetric(matrix):
                    form = 6
                else:
                    form = 1
            else:
                form = 2
        if np.iscomplexobj(vals):
            mtype = 4
            multiplier = 2
        else:
            mtype = 2
            multiplier = 1
        return rows, cols, form, mtype, multiplier

    def _write_ascii_header(self, f, name, matrix, digits, bigmat, form):
        """
        Utility routine that writes the header for ascii matrices.

        Parameters
        ----------
        f : file handle
            Output of open() using binary mode.
        name : string
            Name of matrix.
        matrix : matrix
            Matrix to write.
        digits : integer
            Number of significant digits after the decimal to include
            in the ascii output.
        bigmat : bool
            If true, matrix is to be written in 'bigmat' format.
        form : integer or None
            The matrix form. If None, the form will be determined
            automatically.

        Returns
        -------
        tuple: (cols, multiplier, perline, numlen, numform)
            cols : integer
                Number of columns in matrix.
            multiplier : integer
                2 for complex, 1 for real.
            perline : integer
                Number of values written per row.
            numlen : integer
                Number of characters per value.
            numform : string
                Format string for numbers, eg: '%16.9E'.
        """
        numlen = digits + 5 + self._expdigits  # -1.digitsE-009
        perline = 80 // numlen

        (rows, cols, form, mtype, multiplier) = OP4._get_header_info(matrix, form)

        if bigmat:
            if rows < self._rows4bigmat:
                rows = -rows
        f.write(
            f"{cols:8}{rows:8}{form:8}{mtype:8}{name.upper():8s}"
            f"1P,{perline}E{numlen}.{digits}\n"
        )
        numform = f"%{numlen}.{digits}E"
        return cols, multiplier, perline, numlen, numform

    def _write_ascii(self, f, name, matrix, digits, form):
        """
        Write a matrix to a file in ascii, non-sparse format.

        Parameters
        ----------
        f : file handle
            Output of open() using text mode.
        name : string
            Name of matrix.
        matrix : matrix
            Matrix to write.
        digits : integer
            Number of significant digits after the decimal to include
            in the ascii output.
        form : integer or None
            The matrix form. If None, the form will be determined
            automatically.
        """
        (cols, multiplier, perline, numlen, numform) = self._write_ascii_header(
            f, name, matrix, digits, bigmat=False, form=form
        )

        def _write_col_data(f, v, c, s, elems, perline, numform):
            f.write(f"{c + 1:8}{s + 1:8}{elems:8}\n")
            neven = ((elems - 1) // perline) * perline
            for i in range(0, neven, perline):
                for j in range(perline):
                    f.write(numform % v[i + j])
                f.write("\n")
            for i in range(neven, elems):
                f.write(numform % v[i])
            f.write("\n")

        if isinstance(matrix, np.ndarray):
            for c in range(cols):
                v = matrix[:, c]
                if np.any(v):
                    pv = np.nonzero(v)[0]
                    s = pv[0]
                    e = pv[-1]
                    elems = (e - s + 1) * multiplier
                    v = np.asarray(v[s : e + 1]).ravel()
                    v.dtype = float
                    _write_col_data(f, v, c, s, elems, perline, numform)
        else:
            # sparse matrix:
            rs, cs, vs, cols_with_data = OP4._sparse_sort(matrix)
            dt = float if multiplier == 1 else complex
            for c in cols_with_data:
                pv = (cs == c).nonzero()[0]  # find data for column c
                s = rs[pv[0]]  # first row with value
                e = rs[pv[-1]]  # last row with value
                elems = e - s + 1
                vec = np.zeros(elems, dt)
                vec[rs[pv] - s] = vs[pv]
                elems *= multiplier
                vec.dtype = float
                _write_col_data(f, vec, c, s, elems, perline, numform)
        f.write(f"{cols + 1:8}{1:8}{1:8}\n")
        f.write(numform % 2 ** 0.5)
        f.write("\n")

    @staticmethod
    def _write_ascii_sparse(
        f,
        matrix,
        cols,
        _write_col_header,
        _write_data_string,
        multiplier,
        perline,
        numform,
    ):
        if isinstance(matrix, np.ndarray):
            for c in range(cols):
                v = matrix[:, c]
                if np.any(v):
                    v = np.asarray(v).ravel()
                    ind = OP4._sparse_col_stats(v.nonzero()[0])
                    _write_col_header(f, ind, c, multiplier)
                    for r0, r1 in ind:
                        string = v[r0 : r0 + r1]
                        string.dtype = float
                        _write_data_string(
                            f, string, r0, r1, multiplier, perline, numform
                        )
        else:
            # sparse matrix:
            rs, cs, vs, cols_with_data = OP4._sparse_sort(matrix)
            for c in cols_with_data:
                pv = (cs == c).nonzero()[0]  # find data for column c
                ind = OP4._sparse_col_stats(rs[pv])
                _write_col_header(f, ind, c, multiplier)
                coldata = vs[pv]
                j = 0
                for r0, r1 in ind:
                    string = coldata[j : j + r1]
                    j += r1
                    string.dtype = float
                    _write_data_string(f, string, r0, r1, multiplier, perline, numform)
        f.write(f"{cols + 1:8}{1:8}{1:8}\n")
        f.write(numform % 2 ** 0.5)
        f.write("\n")

    def _write_ascii_nonbigmat(self, f, name, matrix, digits, form):
        """
        Write a matrix to a file in ascii, non-bigmat sparse format.

        Parameters
        ----------
        f : file handle
            Output of open() using binary mode.
        name : string
            Name of matrix.
        matrix : matrix
            Matrix to write.
        digits : integer
            Number of significant digits after the decimal to include
            in the ascii output.
        form : integer or None
            The matrix form. If None, the form will be determined
            automatically.

        Note: if rows > 65535, bigmat is turned on and the
        :func:`write_ascii_bigmat` function is called ...
        that's a Nastran rule.
        """
        if isinstance(matrix, tuple):
            rows, cols = matrix[0].shape
        else:
            rows, cols = matrix.shape

        if rows >= self._rows4bigmat:
            self._write_ascii_bigmat(f, name, matrix, digits, form)
            return
        (cols, multiplier, perline, numlen, numform) = self._write_ascii_header(
            f, name, matrix, digits, bigmat=False, form=form
        )

        def _write_col_header(f, ind, c, multiplier):
            nwords = ind.shape[0] + 2 * sum(ind[:, 1]) * multiplier
            f.write(f"{c + 1:8}{0:8}{nwords:8}\n")

        def _write_data_string(f, string, r0, r1, multiplier, perline, numform):
            L = r1 * 2 * multiplier
            IS = (r0 + 1) + ((L + 1) << 16)
            f.write(f"{IS:12}\n")
            elems = L // 2
            neven = ((elems - 1) // perline) * perline
            for i in range(0, neven, perline):
                for j in range(perline):
                    f.write(numform % string[i + j])
                f.write("\n")
            for i in range(neven, elems):
                f.write(numform % string[i])
            f.write("\n")

        OP4._write_ascii_sparse(
            f,
            matrix,
            cols,
            _write_col_header,
            _write_data_string,
            multiplier,
            perline,
            numform,
        )

    def _write_ascii_bigmat(self, f, name, matrix, digits, form):
        """
        Write a matrix to a file in ascii, bigmat sparse format.

        Parameters
        ----------
        f : file handle
            Output of open() using binary mode.
        name : string
            Name of matrix.
        matrix : matrix
            Matrix to write.
        digits : integer
            Number of significant digits after the decimal to include
            in the ascii output.
        form : integer or None
            The matrix form. If None, the form will be determined
            automatically.
        """
        (cols, multiplier, perline, numlen, numform) = self._write_ascii_header(
            f, name, matrix, digits, bigmat=True, form=form
        )

        def _write_col_header(f, ind, c, multiplier):
            nwords = 2 * ind.shape[0] + 2 * sum(ind[:, 1]) * multiplier
            f.write(f"{c + 1:8}{0:8}{nwords:8}\n")

        def _write_data_string(f, string, r0, r1, multiplier, perline, numform):
            L = r1 * 2 * multiplier
            f.write(f"{L + 1:8}{r0 + 1:8}\n")
            elems = L // 2
            neven = ((elems - 1) // perline) * perline
            for i in range(0, neven, perline):
                for j in range(perline):
                    f.write(numform % string[i + j])
                f.write("\n")
            for i in range(neven, elems):
                f.write(numform % string[i])
            f.write("\n")

        OP4._write_ascii_sparse(
            f,
            matrix,
            cols,
            _write_col_header,
            _write_data_string,
            multiplier,
            perline,
            numform,
        )

    def _write_binary_header(self, f, name, matrix, endian, bigmat, form):
        """
        Utility routine that writes the header for binary matrices.

        Parameters
        ----------
        f : file handle
            Output of open() using binary mode.
        name : string
            Name of matrix.
        matrix : matrix
            Matrix to write.
        endian : string
            Endian setting for binary output:  '' for native, '>' for
            big-endian and '<' for little-endian.
        bigmat : bool
            If true, matrix is to be written in 'bigmat' format.
        form : integer or None
            The matrix form. If None, the form will be determined
            automatically.

        Returns
        -------
        tuple: (cols, multiplier)
            cols : integer
                Number of columns in matrix.
            multiplier : integer
                2 for complex, 1 for real.
        """
        (rows, cols, form, mtype, multiplier) = OP4._get_header_info(matrix, form)

        # write 1st record (24 bytes: 4 4-byte ints, 1 8-byte string)
        name = (f"{name.upper():<8}").encode()
        if bigmat:
            if rows < self._rows4bigmat:
                rows = -rows
        f.write(struct.pack(endian + "5i8si", 24, cols, rows, form, mtype, name, 24))
        return cols, multiplier

    def _write_binary(self, f, name, matrix, endian, form):
        """
        Write a matrix to a file in double precision binary format.

        Parameters
        ----------
        f : file handle
            Output of open() using binary mode.
        name : string
            Name of matrix.
        matrix : matrix
            Matrix to write.
        endian : string
            Endian setting for binary output:  '' for native, '>' for
            big-endian and '<' for little-endian.
        form : integer or None
            The matrix form. If None, the form will be determined
            automatically.
        """
        cols, multiplier = self._write_binary_header(
            f, name, matrix, endian, bigmat=False, form=form
        )

        def _write_col_data(f, v, c, s, elems, endian, colHeader, colTrailer):
            reclen = 3 * 4 + elems * 8
            f.write(colHeader.pack(reclen, c + 1, s + 1, 2 * elems))
            f.write(struct.pack(endian + ("%dd" % elems), *v))
            f.write(colTrailer.pack(reclen))

        colHeader = struct.Struct(endian + "4i")
        colTrailer = struct.Struct(endian + "i")
        if isinstance(matrix, np.ndarray):
            for c in range(cols):
                v = matrix[:, c]
                if np.any(v):
                    pv = np.nonzero(v)[0]
                    s = pv[0]
                    e = pv[-1]
                    elems = (e - s + 1) * multiplier
                    v = np.asarray(v[s : e + 1]).ravel()
                    v.dtype = float
                    _write_col_data(f, v, c, s, elems, endian, colHeader, colTrailer)
        else:
            # sparse matrix:
            rs, cs, vs, cols_with_data = OP4._sparse_sort(matrix)
            dt = float if multiplier == 1 else complex
            for c in cols_with_data:
                pv = (cs == c).nonzero()[0]  # find data for column c
                s = rs[pv[0]]  # first row with value
                e = rs[pv[-1]]  # last row with value
                elems = e - s + 1
                vec = np.zeros(elems, dt)
                vec[rs[pv] - s] = vs[pv]
                elems *= multiplier
                vec.dtype = float
                _write_col_data(f, vec, c, s, elems, endian, colHeader, colTrailer)
        reclen = 3 * 4 + 8
        f.write(colHeader.pack(reclen, cols + 1, 1, 2))
        f.write(struct.pack(endian + "d", 2 ** 0.5))
        f.write(colTrailer.pack(reclen))

    @staticmethod
    def _write_binary_sparse(
        f,
        matrix,
        cols,
        _write_col_header,
        _write_data_string,
        multiplier,
        colHeader,
        colTrailer,
        LrStruct,
        endian,
    ):
        if isinstance(matrix, np.ndarray):
            for c in range(cols):
                v = matrix[:, c]
                if np.any(v):
                    v = np.asarray(v).ravel()
                    ind = OP4._sparse_col_stats(v.nonzero()[0])
                    reclen = _write_col_header(f, ind, c, multiplier, colHeader)
                    for r0, r1 in ind:
                        string = v[r0 : r0 + r1]
                        string.dtype = float
                        _write_data_string(
                            f, string, r0, r1, multiplier, LrStruct, endian
                        )
                    f.write(colTrailer.pack(reclen))
        else:
            # sparse matrix:
            rs, cs, vs, cols_with_data = OP4._sparse_sort(matrix)
            for c in cols_with_data:
                pv = (cs == c).nonzero()[0]  # find data for column c
                ind = OP4._sparse_col_stats(rs[pv])
                reclen = _write_col_header(f, ind, c, multiplier, colHeader)
                coldata = vs[pv]
                j = 0
                for r0, r1 in ind:
                    string = coldata[j : j + r1]
                    j += r1
                    string.dtype = float
                    _write_data_string(f, string, r0, r1, multiplier, LrStruct, endian)
                f.write(colTrailer.pack(reclen))
        reclen = 3 * 4 + 8
        f.write(colHeader.pack(reclen, cols + 1, 1, 2))
        f.write(struct.pack(endian + "d", 2 ** 0.5))
        f.write(colTrailer.pack(reclen))

    def _write_binary_nonbigmat(self, f, name, matrix, endian, form):
        """
        Write a matrix to a file in double precision binary,
        non-bigmat sparse format.

        Parameters
        ----------
        f : file handle
            Output of open() using binary mode.
        name : string
            Name of matrix.
        matrix : matrix
            Matrix to write.
        endian : string
            Endian setting for binary output:  '' for native, '>' for
            big-endian and '<' for little-endian.
        form : integer or None
            The matrix form. If None, the form will be determined
            automatically.

        Note: if rows > 65535, bigmat is turned on and the
        :func:`write_binary_bigmat` function is called ...
        that's a Nastran rule.
        """
        if isinstance(matrix, tuple):
            rows, cols = matrix[0].shape
        else:
            rows, cols = matrix.shape

        if rows >= self._rows4bigmat:
            self._write_binary_bigmat(f, name, matrix, endian, form)
            return

        cols, multiplier = self._write_binary_header(
            f, name, matrix, endian, bigmat=False, form=form
        )
        colHeader = struct.Struct(endian + "4i")
        colTrailer = struct.Struct(endian + "i")

        def _write_col_header(f, ind, c, multiplier, colHeader):
            nwords = ind.shape[0] + 2 * sum(ind[:, 1]) * multiplier
            reclen = (3 + nwords) * 4
            f.write(colHeader.pack(reclen, c + 1, 0, nwords))
            return reclen

        def _write_data_string(f, string, r0, r1, multiplier, colTrailer, endian):
            L = r1 * 2 * multiplier
            IS = (r0 + 1) + ((L + 1) << 16)
            f.write(colTrailer.pack(IS))
            f.write(struct.pack(endian + ("%dd" % len(string)), *string))

        OP4._write_binary_sparse(
            f,
            matrix,
            cols,
            _write_col_header,
            _write_data_string,
            multiplier,
            colHeader,
            colTrailer,
            colTrailer,
            endian,
        )

    def _write_binary_bigmat(self, f, name, matrix, endian, form):
        """
        Write a matrix to a file in double precision binary, bigmat
        sparse format.

        Parameters
        ----------
        f : file handle
            Output of open() using binary mode.
        name : string
            Name of matrix.
        matrix : matrix
            Matrix to write.
        endian : string
            Endian setting for binary output:  '' for native, '>' for
            big-endian and '<' for little-endian.
        form : integer or None
            The matrix form. If None, the form will be determined
            automatically.
        """
        cols, multiplier = self._write_binary_header(
            f, name, matrix, endian, bigmat=True, form=form
        )
        colHeader = struct.Struct(endian + "4i")
        colTrailer = struct.Struct(endian + "i")
        LrStruct = struct.Struct(endian + "ii")

        def _write_col_header(f, ind, c, multiplier, colHeader):
            nwords = 2 * ind.shape[0] + 2 * sum(ind[:, 1]) * multiplier
            reclen = (3 + nwords) * 4
            f.write(colHeader.pack(reclen, c + 1, 0, nwords))
            return reclen

        def _write_data_string(f, string, r0, r1, multiplier, LrStruct, endian):
            L = r1 * 2 * multiplier
            f.write(LrStruct.pack(L + 1, r0 + 1))
            f.write(struct.pack(endian + ("%dd" % len(string)), *string))

        OP4._write_binary_sparse(
            f,
            matrix,
            cols,
            _write_col_header,
            _write_data_string,
            multiplier,
            colHeader,
            colTrailer,
            LrStruct,
            endian,
        )

    def dctload(self, filename, namelist=None, justmatrix=False, sparse=False):
        """
        Read all matching matrices from op4 file into dictionary.

        Parameters
        ----------
        filename : string
            Name of op4 file to read.
        namelist : list, string, or None; optional
            List of variable names to read in, or string with name of
            the single variable to read in, or None. If None, all
            matrices are read in.
        justmatrix : bool; optional
            If True, only the matrix is stored in the dictionary. If
            False, a tuple of ``(matrix, form, mtype)`` is stored.
        sparse : bool or None or two-tuple_like; optional
            Specifies whether output matrices will be regular numpy
            arrays or sparse arrays. If not two-tuple_like:

            ========   ===============================================
            `sparse`   Action
            ========   ===============================================
             None      Auto setting: each matrix will be sparse if and
                       only if it was written in a sparse format
             True      Matrices will be returned in sparse format
             False     Matrices will be returned in regular (dense)
                       numpy arrays
            ========   ===============================================

            If `sparse` is two-tuple_like, the first element is either
            None, True, or False (see table above) and the second
            element is a callable, as in: ``X = callable(X)``. A
            common usage of the callable would be to convert from
            "COO" sparse form (see :class:`scipy.sparse.coo_matrix`)
            to a more desirable form. For example, to ensure *all*
            matrices are returned in CSC form (see
            :class:`scipy.sparse.csc_matrix`) use::

                sparse=(True, scipy.sparse.coo_matrix.tocsc)

            The callable is ignored for non-sparse matrices.

        Returns
        -------
        dct : :class:`collections.OrderedDict`
            Keys are the lower-case matrix names and the values are
            either just the matrix or a tuple of:
            ``(matrix, form, mtype)`` depending on `justmatrix`.

        Notes
        -----
        The default form for sparse matrices is the "COO" sparse form
        (see :class:`scipy.sparse.coo_matrix`). To override, provide a
        callable in the `sparse` option (see above).

        See also
        --------
        :func:`listload`, :func:`write` (or :func:`save`),
        :func:`dir`.
        """
        if isinstance(namelist, str):
            namelist = [namelist]
        self._op4open_read(filename)
        dct = collections.OrderedDict()
        try:
            if self._ascii:
                loadfunc = self._loadop4_ascii
            else:
                loadfunc = self._loadop4_binary
            while 1:
                name, X, form, mtype = loadfunc(patternlist=namelist, sparse=sparse)
                if not name:
                    break
                if justmatrix:
                    dct[name] = X
                else:
                    dct[name] = X, form, mtype
        finally:
            self._op4close()
        return dct

    def listload(self, filename, namelist=None, sparse=False):
        """
        Read all matching matrices from op4 file into a list; useful
        if op4 file has duplicate names.

        Parameters
        ----------
        filename : string
            Name of op4 file to read.
        namelist : list, string, or None; optional
            List of variable names to read in, or string with name of
            the single variable to read in, or None. If None, all
            matrices are read in.
        sparse : bool or None or two-tuple_like; optional
            Specifies whether output matrices will be regular numpy
            arrays or sparse arrays. If not two-tuple_like:

            ========   ===============================================
            `sparse`   Action
            ========   ===============================================
             None      Auto setting: each matrix will be sparse if and
                       only if it was written in a sparse format
             True      Matrices will be returned in sparse format
             False     Matrices will be returned in regular (dense)
                       numpy arrays
            ========   ===============================================

            If `sparse` is two-tuple_like, the first element is either
            None, True, or False (see table above) and the second
            element is a callable, as in: ``X = callable(X)``. A
            common usage of the callable would be to convert from
            "COO" sparse form (see :class:`scipy.sparse.coo_matrix`)
            to a more desirable form. For example, to ensure *all*
            matrices are returned in CSC form (see
            :class:`scipy.sparse.csc_matrix`) use::

                sparse=(True, scipy.sparse.coo_matrix.tocsc)

            The callable is ignored for non-sparse matrices.

        Returns
        -------
        names : list
            Lower-case list of matrix names in order as read.
        matrices : list
            List of matrices in order as read.
        forms : list
            List of integers specifying the Nastran form of each
            matrix.
        mtypes : list
            List of integers specifying the Nastran type of each
            matrix.

        Notes
        -----
        The default form for sparse matrices is the "COO" sparse form
        (see :class:`scipy.sparse.coo_matrix`). To override, provide a
        callable in the `sparse` option (see above).

        See also
        --------
        :func:`dctload`, :func:`write` (or :func:`save`), :func:`dir`.
        """
        if isinstance(namelist, str):
            namelist = [namelist]
        self._op4open_read(filename)
        names = []
        matrices = []
        forms = []
        mtypes = []
        try:
            if self._ascii:
                loadfunc = self._loadop4_ascii
            else:
                loadfunc = self._loadop4_binary
            while 1:
                name, X, form, mtype = loadfunc(patternlist=namelist, sparse=sparse)
                if not name:
                    break
                names.append(name)
                matrices.append(X)
                forms.append(form)
                mtypes.append(mtype)
        finally:
            self._op4close()
        return names, matrices, forms, mtypes

    def load(self, filename, namelist=None, into="dct", justmatrix=False, sparse=False):
        """
        Read all matching matrices from op4 file into dictionary or
        list; interface to :func:`dctload` and :func:`listload`.

        Parameters
        ----------
        filename : string
            Name of op4 file to read.
        namelist : list, string, or None; optional
            List of variable names to read in, or string with name of
            the single variable to read in, or None. If None, all
            matrices are read in.
        into : string; optional
            Either 'dct' or 'list'. Use 'list' if multiple matrices
            share the same name. See below.
        justmatrix : bool; optional
            If True, only the matrix is stored in the dictionary. If
            False, a tuple of ``(matrix, form, mtype)`` is stored.
            This option is ignored if ``into == 'list'``.
        sparse : bool or None or two-tuple_like; optional
            Specifies whether output matrices will be regular numpy
            arrays or sparse arrays. If not two-tuple_like:

            ========   ===============================================
            `sparse`   Action
            ========   ===============================================
             None      Auto setting: each matrix will be sparse if and
                       only if it was written in a sparse format
             True      Matrices will be returned in sparse format
             False     Matrices will be returned in regular (dense)
                       numpy arrays
            ========   ===============================================

            If `sparse` is two-tuple_like, the first element is either
            None, True, or False (see table above) and the second
            element is a callable, as in: ``X = callable(X)``. A
            common usage of the callable would be to convert from
            "COO" sparse form (see :class:`scipy.sparse.coo_matrix`)
            to a more desirable form. For example, to ensure *all*
            matrices are returned in CSC form (see
            :class:`scipy.sparse.csc_matrix`) use::

                sparse=(True, scipy.sparse.coo_matrix.tocsc)

            The callable is ignored for non-sparse matrices.

        Returns
        -------
        dct : :class:`collections.OrderedDict`, if ``into == 'dct'``
            Keys are the lower-case matrix names and the values are
            either just the matrix or a tuple of:
            ``(matrix, form, mtype)`` depending on `justmatrix`.

        tup : tuple, if ``into == 'list'``
            Tuple of 4 lists: ``(names, matrices, forms, mtypes)``

        Notes
        -----
        The default form for sparse matrices is the "COO" sparse form
        (see :class:`scipy.sparse.coo_matrix`). To override, provide a
        callable in the `sparse` option (see above).

        See also
        --------
        :func:`write` (or :func:`save`), :func:`dir`.
        """
        if into == "dct":
            return self.dctload(filename, namelist, justmatrix, sparse)
        elif into == "list":
            return self.listload(filename, namelist, sparse)
        raise ValueError('invalid "into" option')

    def dir(self, filename, verbose=True):
        """
        Directory of all matrices in op4 file.

        Parameters
        ----------
        filename : string
            Name of op4 file to read.
        verbose : bool; optional
            If true, directory will be printed to screen.

        Returns
        -------
        names : list
            Lower-case list of matrix names in order as read.
        sizes : list
            List of sizes [(r1, c1), (r2, c2), ...], for each
            matrix.
        forms : list
            List of integers specifying the Nastran form of each
            matrix.
        mtypes : list
            List of integers specifying the Nastran type of each
            matrix.

        See also
        --------
        :func:`dctload`, :func:`listload`, :func:`write` (or
        :func:`save`).
        """
        self._op4open_read(filename)
        names = []
        sizes = []
        forms = []
        mtypes = []
        try:
            if self._ascii:
                loadfunc = self._loadop4_ascii
            else:
                loadfunc = self._loadop4_binary
            while 1:
                name, X, form, mtype = loadfunc(listonly=True)
                if not name:
                    break
                names.append(name)
                sizes.append(X)
                forms.append(form)
                mtypes.append(mtype)
            if verbose:
                for n, s, f, m in zip(names, sizes, forms, mtypes):
                    print(f"{n:8}, {s[0]:6} x {s[1]:<6}, form={f}, mtype={m}")
        finally:
            self._op4close()
        return names, sizes, forms, mtypes

    def write(
        self,
        filename,
        names,
        matrices=None,
        binary=True,
        digits=16,
        endian="",
        sparse="auto",
        forms=None,
    ):
        """
        Write op4 file.

        Parameters
        ----------
        filename : string
            Name of file.
        names : dictionary/OrderedDict or list or string
            Dictionary indexed by the matrix names with the values
            being either the matrices or a tuple/list where the first
            two items are ``(matrix, form)``. Alternatively, `names`
            can also be a list of matrix names (strings) or a single
            name (string) if just one matrix is to be saved. If using
            the list inputs, `matrices` is required to specify the
            matrices and, if `forms` needs to specified, that input
            would also be required.
        matrices : array or list; optional
            2d ndarray or list of 2d ndarrays. Ignored if `names` is
            a dictionary. Same length as `names`.
        binary : bool; optional
            If true, a double precision binary file is written;
            otherwise an ascii file is created.
        digits : integer; optional
            Number of significant digits after the decimal to include
            in the ascii output. Ignored for binary files.
        endian : string; optional
            Endian setting for binary output:  '' for native, '>' for
            big-endian and '<' for little-endian.
        sparse : string; optional
            Specifies the output format:

            ===========   ===========================================
             `sparse`     Action
            ===========   ===========================================
            'auto'        Each 2d ndarray will be written in "dense"
                          format and each sparse matrix will be
                          written in "bigmat" sparse format
            'bigmat'      Each matrix (whether sparse or not) is
                          written in "bigmat" sparse format
            'dense'       Each matrix will be written in "dense"
                          format
            'nonbigmat'   Each matrix is written in "nonbigmat"
                          format. Note that if the number of rows is
                          > 65535, then the "bigmat" format is used.
            ===========   ===========================================

        forms : integer or list or None; optional
            The matrix form(s). If None, the forms will be determined
            automatically to be 1, 2, or 6. Ignored if `names` is a
            dictionary (which would contain this information). If not
            None, `forms` must be the same length as `names`, but you
            can use None as the form for one or more matrices (see
            example in :func:`pyyeti.nastran.op4.write`). From Nastran
            documentation:

            ======   ==============================
             form    Matrix format
            ======   ==============================
               1     Square
               2     Rectangular
               3     Diagonal
               4     Lower triangular factor
               5     Upper triangular factor
               6     Symmetric
               8     Identity
               9     Pseudo identity
              10     Cholesky factor
              11     Trapezoidal factor
              13     Sparse lower triangular factor
              15     Sparse upper triangular factor
            ======   ==============================

            .. warning::
                The validity of the values in `forms` is not checked
                in any way.

        Returns
        -------
        None.

        Notes
        -----
        To write multiple matrices that have the same name, `names`
        must be a list, not a dictionary. If a list, the order is
        maintained. If a dictionary, the matrices are written in the
        order they are retrieved from the dictionary; use a
        :class:`collections.OrderedDict` to specify a certain order.

        See the examples in :func:`pyyeti.nastran.op4.write`.

        See also
        --------
        :func:`pyyeti.nastran.op4.write`, :func:`dctload`,
        :func:`listload`, :func:`dir`.
        """

        def ensure_dp(m):
            """
            Ensure double precision values
            """
            if np.iscomplexobj(m):
                if m.dtype != np.complex128:
                    return m.astype(np.complex128)
            elif m.dtype != np.float64:
                return m.astype(np.float64)
            return m

        # ensure double precision 2d arrays or tuple as returned by
        # scipy.sparse.find:
        def ensure_2d_dp(m):
            """
            Ensures 2d double precision array or tuple as returned by
            scipy.sparse.find
            """
            if sp.issparse(m):
                # return m
                i, j, v = sp.find(m)
                return m, i, j, ensure_dp(v)
            m = np.atleast_2d(m)
            if m.ndim > 2:
                raise ValueError("found array with greater than 2 dimensions.")
            return ensure_dp(m)

        if isinstance(names, collections.Mapping):
            _names = []
            matrices = []
            forms = []
            for nm, val in names.items():
                if isinstance(val, (list, tuple)):
                    _names.append(nm)
                    matrices.append(val[0])
                    forms.append(val[1])
                else:
                    _names.append(nm)
                    matrices.append(val)
                    forms.append(None)
            names = _names
        else:
            if not isinstance(names, (list, tuple)):
                names = [names]
            if not isinstance(matrices, (list, tuple)):
                matrices = [matrices]
            if forms is not None and not isinstance(forms, (list, tuple)):
                forms = [forms]

        if forms is None:
            forms = [None] * len(names)

        matrices = [ensure_2d_dp(matrix) for matrix in matrices]

        if binary:
            if sparse == "dense":
                wrtfunc = self._write_binary
            elif sparse == "bigmat":
                wrtfunc = self._write_binary_bigmat
            elif sparse == "nonbigmat":
                wrtfunc = self._write_binary_nonbigmat
            elif sparse != "auto":
                raise ValueError("invalid sparse option")
            with open(filename, "wb") as f:
                for name, matrix, form in zip(names, matrices, forms):
                    if sparse == "auto":
                        if isinstance(matrix, tuple):
                            wrtfunc = self._write_binary_bigmat
                        else:
                            wrtfunc = self._write_binary
                    wrtfunc(f, name, matrix, endian, form)
        else:
            if sparse == "dense":
                wrtfunc = self._write_ascii
            elif sparse == "bigmat":
                wrtfunc = self._write_ascii_bigmat
            elif sparse == "nonbigmat":
                wrtfunc = self._write_ascii_nonbigmat
            elif sparse != "auto":
                raise ValueError("invalid sparse option")
            with open(filename, "w") as f:
                for name, matrix, form in zip(names, matrices, forms):
                    if sparse == "auto":
                        if isinstance(matrix, tuple):
                            wrtfunc = self._write_ascii_bigmat
                        else:
                            wrtfunc = self._write_ascii
                    wrtfunc(f, name, matrix, digits, form)


def load(filename=None, namelist=None, into="dct", justmatrix=False, sparse=False):
    """
    Read all matching matrices from op4 file into dictionary or list;
    non-member version of :func:`OP4.load`.

    This is a the same as :func:`read` except `justmatrix` default is
    False.

    Parameters
    ----------
    filename : string or None; optional
        Name of op4 file to read. Can also be the name of a directory
        or None; in these cases, a GUI is opened for file selection.
    namelist : list, string, or None; optional
        List of variable names to read in, or string with name of the
        single variable to read in, or None. If None, all matrices
        are read in.
    into : string; optional
        Either 'dct' or 'list'. Use 'list' if multiple matrices share
        the same name. See below.
    justmatrix : bool; optional
        If True, only the matrix is stored in the dictionary. If
        False, a tuple of ``(matrix, form, mtype)`` is stored.
        This option is ignored if ``into == 'list'``.
    sparse : bool or None or two-tuple_like; optional
        Specifies whether output matrices will be regular numpy arrays
        or sparse arrays. If not two-tuple_like:

        ========   ===============================================
        `sparse`   Action
        ========   ===============================================
         None      Auto setting: each matrix will be sparse if and
                   only if it was written in a sparse format
         True      Matrices will be returned in sparse format
         False     Matrices will be returned in regular (dense)
                   numpy arrays
        ========   ===============================================

        If `sparse` is two-tuple_like, the first element is either
        None, True, or False (see table above) and the second element
        is a callable, as in: ``X = callable(X)``. A common usage of
        the callable would be to convert from "COO" sparse form (see
        :class:`scipy.sparse.coo_matrix`) to a more desirable
        form. For example, to ensure *all* matrices are returned in
        CSC form (see :class:`scipy.sparse.csc_matrix`) use::

            sparse=(True, scipy.sparse.coo_matrix.tocsc)

        The callable is ignored for non-sparse matrices.

    Returns
    -------
    dct : :class:`collections.OrderedDict`, if ``into == 'dct'``
        Keys are the lower-case matrix names and the values are
        either just the matrix or a tuple of:
        ``(matrix, form, mtype)`` depending on `justmatrix`.

    tup : tuple, if ``into == 'list'``
        Tuple of 4 lists: ``(names, matrices, forms, mtypes)``

    Notes
    -----
    The default form for sparse matrices is the "COO" sparse form (see
    :class:`scipy.sparse.coo_matrix`). To override, provide a callable
    in the `sparse` option (see above).

    Examples
    --------
    This examples translates a sparse format binary op4 file to a
    simpler ascii format while preserving the matrix forms.

    First, create a file in "bigmat" sparse format and set the form on
    the "m" matrix to be symmetric (form=6):

    >>> import numpy as np
    >>> from pyyeti.nastran import op4
    >>> m = np.array([[1, 2], [2.1, 3]])
    >>> k = np.array([3, 5])
    >>> b = np.array([4, 6])
    >>> names = ['m', 'k', 'b']
    >>> values = [eval(n) for n in names]
    >>> op4.write('mkb.op4', names, values, forms=[6, 2, 2],
    ...           sparse='bigmat')

    Now, translate it to simple ascii, preserving the forms:

    >>> dct = op4.load('mkb.op4')
    >>> op4.write('mkb_ascii.op4', dct, binary=False)

    Check that the order and forms are the same:

    >>> _ = op4.dir('mkb.op4')
    m       ,      2 x 2     , form=6, mtype=2
    k       ,      1 x 2     , form=2, mtype=2
    b       ,      1 x 2     , form=2, mtype=2

    >>> _ = op4.dir('mkb_ascii.op4')
    m       ,      2 x 2     , form=6, mtype=2
    k       ,      1 x 2     , form=2, mtype=2
    b       ,      1 x 2     , form=2, mtype=2

    Clean up:

    >>> import os
    >>> os.remove('mkb.op4')
    >>> os.remove('mkb_ascii.op4')

    See also
    --------
    :func:`read`, :func:`write` (or :func:`save`), :func:`dir`.
    """
    filename = guitools.get_file_name(filename, read=True)
    if into == "dct":
        return OP4().dctload(filename, namelist, justmatrix, sparse)
    elif into == "list":
        return OP4().listload(filename, namelist, sparse)
    raise ValueError('invalid "into" option')


def read(filename=None, namelist=None, into="dct", justmatrix=True, sparse=False):
    """
    Read all matching matrices from op4 file into dictionary or list;
    non-member version of :func:`OP4.load`.

    This is a the same as :func:`load` except `justmatrix` default is
    True.

    Parameters
    ----------
    filename : string or None; optional
        Name of op4 file to read. Can also be the name of a directory
        or None; in these cases, a GUI is opened for file selection.
    namelist : list, string, or None; optional
        List of variable names to read in, or string with name of the
        single variable to read in, or None. If None, all matrices
        are read in.
    into : string; optional
        Either 'dct' or 'list'. Use 'list' if multiple matrices share
        the same name. See below.
    justmatrix : bool; optional
        If True, only the matrix is stored in the dictionary. If
        False, a tuple of ``(matrix, form, mtype)`` is stored.
        This option is ignored if ``into == 'list'``.
    sparse : bool or None or two-tuple_like; optional
        Specifies whether output matrices will be regular numpy arrays
        or sparse arrays. If not two-tuple_like:

        ========   ===============================================
        `sparse`   Action
        ========   ===============================================
         None      Auto setting: each matrix will be sparse if and
                   only if it was written in a sparse format
         True      Matrices will be returned in sparse format
         False     Matrices will be returned in regular (dense)
                   numpy arrays
        ========   ===============================================

        If `sparse` is two-tuple_like, the first element is either
        None, True, or False (see table above) and the second element
        is a callable, as in: ``X = callable(X)``. A common usage of
        the callable would be to convert from "COO" sparse form (see
        :class:`scipy.sparse.coo_matrix`) to a more desirable
        form. For example, to ensure *all* matrices are returned in
        CSC form (see :class:`scipy.sparse.csc_matrix`) use::

            sparse=(True, scipy.sparse.coo_matrix.tocsc)

        The callable is ignored for non-sparse matrices.

    Returns
    -------
    dct : :class:`collections.OrderedDict`, if ``into == 'dct'``
        Keys are the lower-case matrix names and the values are
        either just the matrix or a tuple of:
        ``(matrix, form, mtype)`` depending on `justmatrix`.

    tup : tuple, if ``into == 'list'``
        Tuple of 4 lists: ``(names, matrices, forms, mtypes)``

    Notes
    -----
    The default form for sparse matrices is the "COO" sparse form (see
    :class:`scipy.sparse.coo_matrix`). To override, provide a callable
    in the `sparse` option (see above).

    See also
    --------
    :func:`load`, :func:`write` (or :func:`save`), :func:`dir`.
    """
    return load(filename, namelist, into, justmatrix, sparse)


def dir(filename=None, verbose=True):
    """
    Directory of all matrices in op4 file; non-member version of
    :func:`OP4.dir`.

    Parameters
    ----------
    filename : string or None; optional
        Name of op4 file to read. Can also be the name of a directory
        or None; in these cases, a GUI is opened for file selection.
    verbose : bool; optional
        If true, directory will be printed to screen.

    Returns
    -------
    names : list
        Lower-case list of matrix names in order as read.
    sizes : list
        List of sizes [(r1, c1), (r2, c2), ...], for each
        matrix.
    forms : list
        List of integers specifying the Nastran form of each
        matrix.
    mtypes : list
        List of integers specifying the Nastran type of each
        matrix.

    See also
    --------
    :func:`load`, :func:`write` (or :func:`save`).
    """
    filename = guitools.get_file_name(filename, read=True)
    return OP4().dir(filename, verbose)


def write(
    filename,
    names,
    matrices=None,
    binary=True,
    digits=16,
    endian="",
    sparse="auto",
    forms=None,
):
    """
    Write op4 file; non-member version of :func:`OP4.write`.

    Parameters
    ----------
    filename : string or None
        Name of file. Can also be the name of a directory or None; in
        these cases, a GUI is opened for file selection.
    names : dictionary/OrderedDict or list or string
        Dictionary indexed by the matrix names with the values being
        either the matrices or a tuple/list where the first two items
        are ``(matrix, form)``. Alternatively, `names` can also be a
        list of matrix names (strings) or a single name (string) if
        just one matrix is to be saved. If using the list inputs,
        `matrices` is required to specify the matrices and, if `forms`
        needs to specified, that input would also be required.
    matrices : array or list; optional
        2d ndarray or list of 2d ndarrays. Ignored if `names` is
        a dictionary. Same length as `names`.
    binary : bool; optional
        If true, a double precision binary file is written;
        otherwise an ascii file is created.
    digits : integer; optional
        Number of significant digits after the decimal to include
        in the ascii output. Ignored for binary files.
    endian : string; optional
        Endian setting for binary output:  '' for native, '>' for
        big-endian and '<' for little-endian.
    sparse : string; optional
        Specifies the output format:

        ===========   ===========================================
         `sparse`     Action
        ===========   ===========================================
        'auto'        Each 2d ndarray will be written in "dense"
                      format and each sparse matrix will be
                      written in "bigmat" sparse format
        'bigmat'      Each matrix (whether sparse or not) is
                      written in "bigmat" sparse format
        'dense'       Each matrix will be written in "dense"
                      format
        'nonbigmat'   Each matrix is written in "nonbigmat"
                      format. Note that if the number of rows is
                      > 65535, then the "bigmat" format is used.
        ===========   ===========================================

    forms : integer or list or None; optional
        The matrix form(s). If None, the forms will be determined
        automatically to be 1, 2, or 6. Ignored if `names` is a
        dictionary (which would contain this information). If not
        None, `forms` must be the same length as `names`, but you can
        use None as the form for one or more matrices (see example
        below). From Nastran documentation:

        ======   ==============================
         form    Matrix format
        ======   ==============================
           1     Square
           2     Rectangular
           3     Diagonal
           4     Lower triangular factor
           5     Upper triangular factor
           6     Symmetric
           8     Identity
           9     Pseudo identity
          10     Cholesky factor
          11     Trapezoidal factor
          13     Sparse lower triangular factor
          15     Sparse upper triangular factor
        ======   ==============================

        .. warning::
            The validity of the values in `forms` is not checked in
            any way.

    Returns
    -------
    None.

    Notes
    -----
    To write multiple matrices that have the same name, `names` must
    be a list, not a dictionary. If a list, the order is
    maintained. If a dictionary, the matrices are written in the order
    they are retrieved from the dictionary; use a
    :class:`collections.OrderedDict` to specify a certain order.

    `save` is an alias for `write`.

    Examples
    --------
    To write m, k, b, in that order to a binary file, you could
    use lists or an OrderedDict:

    >>> import numpy as np
    >>> from pyyeti.nastran import op4
    >>> m = np.array([[1, 2], [2, 3]])
    >>> k = np.array([3, 5])
    >>> b = np.array([4, 6])
    >>> names = ['m', 'k', 'b']
    >>> values = [eval(n) for n in names]
    >>> op4.write('mkb.op4', names, values)
    >>> _ = op4.dir('mkb.op4')
    m       ,      2 x 2     , form=6, mtype=2
    k       ,      1 x 2     , form=2, mtype=2
    b       ,      1 x 2     , form=2, mtype=2

    Or, order is also maintained when using an OrderedDict:

    >>> from collections import OrderedDict
    >>> odct = OrderedDict()
    >>> for n in names:
    ...     odct[n] = eval(n)
    >>> op4.write('mkb.op4', odct)
    >>> _ = op4.dir('mkb.op4')
    m       ,      2 x 2     , form=6, mtype=2
    k       ,      1 x 2     , form=2, mtype=2
    b       ,      1 x 2     , form=2, mtype=2

    On the other hand, if you don't care about the order, you could
    use a regular dictionary. (In more recent versions of Python, this
    may behave like an OrderedDict.):

    >>> op4.write('mkb.op4', dict(m=m, k=k, b=b))
    >>> _ = op4.dir('mkb.op4')         # doctest: +SKIP
    m       ,      2 x 2     , form=6, mtype=2
    k       ,      1 x 2     , form=2, mtype=2
    b       ,      1 x 2     , form=2, mtype=2

    To specify the forms, include the `forms` option in either the
    list approach or the dictionary approach. Here, we'll just say
    that "m" is square (not symmetric) for demonstration. The forms
    for "k" and "b" will be automatically detected. First, the list
    approach:

    >>> op4.write('mkb.op4', ['m', 'k', 'b'], [m, k, b],
    ...           forms=[1, None, None])
    >>> _ = op4.dir('mkb.op4')
    m       ,      2 x 2     , form=1, mtype=2
    k       ,      1 x 2     , form=2, mtype=2
    b       ,      1 x 2     , form=2, mtype=2

    Next, the dictionary approach. Since the form is attached to each
    matrix, it is only required if not None:

    >>> odct = OrderedDict()
    >>> odct['m'] = (m, 1)
    >>> odct['k'] = k
    >>> odct['b'] = b
    >>> op4.write('mkb.op4', odct)
    >>> _ = op4.dir('mkb.op4')
    m       ,      2 x 2     , form=1, mtype=2
    k       ,      1 x 2     , form=2, mtype=2
    b       ,      1 x 2     , form=2, mtype=2

    Clean up:

    >>> import os
    >>> os.remove('mkb.op4')

    See also
    --------
    :func:`load`, :func:`dir`.
    """
    filename = guitools.get_file_name(filename, read=False)
    OP4().write(filename, names, matrices, binary, digits, endian, sparse, forms)


# create `save` as an alias for `write`
save = write



# -*- coding: utf-8 -*-
"""
A class for interactively annotating 2D matplotlib plots.

This class was inspired by code posted by HappyLeapSecond which was
derived from code posted by Joe Kington. See:

http://stackoverflow.com/questions/13306519/\
get-data-from-plot-with-matplotlib
"""

import numpy as np
import matplotlib.pyplot as plt


def form1(x, y, n, ind, ax, line):
    """Default annotation for plots with 1 line"""
    return f"x: {x:.5g}\ny: {y:.5g}"


def form2(x, y, n, ind, ax, line):
    """Default annotation for plots with more than 1 line"""
    return f"x: {x:.5g}\ny: {y:.5g}\n{line.get_label()}"


def _ensure_iterable(a):
    if not isinstance(a, (list, tuple)):
        return [a]
    return a


class DataCursor(object):
    r"""
    Class to show x, y data points and to allow selection of points
    for annotations.

    Attributes
    ----------
    hover : bool
        If True, an annotated large green, semi-transparent dot is
        displayed that follows the mouse as long as the mouse is
        inside the axes. Note: setting this directly is possible; you
        just have to turn the DataCursor off and back on for the
        setting to take effect. For example::

            from pyyeti.datacursor import DC
            DC.hover = False
            DC.off()
            DC.on()

    form1 : function
        Function to format the x, y data for the annotation. This is
        called for axes that have only one line. Must accept the four
        inputs listed and and return a string. The argument `n` is the
        line number starting at 0 and `ind` is the index into the data
        vectors. `form1` defaults to::

          def form1(x, y, n, ind, ax, line):
              return f"x: {x:.5g}\ny: {y:5g}"

    form2 : function
        Function to format the x, y data for the annotation. This is
        called for axes that have more than one line. Same signature
        as `form1`. Defaults to::

          def form2(x, y, n, ind, ax, line):
              return f"x: {x:.5g}\ny: {y:.5g}\n{line.get_label()}"

    offsets : tuple
        Two element tuple containing x and y offsets in points for the
        annotation.
    bbox : dict; optional
        Defines the `bbox` parameter for
        :meth:`matplotlib.axes.Axes.annotate`
    arrowprops : dict; optional
        Defines the `arrowprops` parameter for
        :meth:`matplotlib.axes.Axes.annotate`
    followdot : dict; optional
        Typically defines the `s`, `color`, and `alpha` settings (and
        possibly others as desired) for
        :meth:`matplotlib.axes.Axes.scatter`. That function is used
        for drawing the "dot" on the plot that follows the mouse and
        highlights the currently selected data point.
    permdot : dict; optional
        Similar to `followdot` except this is a "permanent" dot;
        this gets placed after left clicking.
    xypoints : list
        Contains [x, y] data pairs for each (non-deleted) left-click.
    inds : list
        Contains index of [x, y] data pairs for each (non-deleted)
        left-click. Same length as `xypoints`.
    lines : list
        Contains line object handles for the selected lines. Same
        length as `xypoints`.
    linenums : list
        Contains line number for each [x, y] data pair, starting at 0.
        Same length as `xypoints`.
    pts : list
        Contains the :func:`matplotlib.pyplot.scatter` object handles
        for the selected points. Same length as `xypoints`.
    notes : list
        Contains the annotation object handles for the selected points.
        Same length as `xypoints`.

    Notes
    -----
    Having multiple data-cursors active at the same time is
    undesirable. Therefore, this module instantiates one DataCursor
    object called `DC` during the initial import. It is recommended to
    always use `DC` rather than instantiating new DataCursor objects.
    From within a script::

        from datacursor import DC
        ...
        DC.getdata()   # blocks until DataCursor is turned off via 't'
        DC.getdata(n)  # blocks until user selects `n` points (or
                       # DataCursor is turned off via 't')

    From an interactive prompt::

        from datacursor import DC
        DC.on()       # doesn't block, but DC is active

    Once the DataCursor is turned on, you'll just mouse over the data
    to see selected data points. These operations are available (when
    the mouse is inside the axes):

    ===========  =====================================================
    Action       Description
    ===========  =====================================================
    left-click   Data point will be stored in the member list
                 `xypoints`. The other members (`inds`, `lines`,
                 `linenums`, `pts` and `notes`) are updated as well.
    right-click  Last point is deleted from `xypoints` and from the
                 other members.
    typing 't'   Turns off DataCursor. To turn on, use ``DC.on``. Note
                 that turning on will reset `xypoints` and the other
                 members.
    typing 'D'   Deletes last point AND removes the line from the plot
                 via ``line_handle.remove()``. Any older annotations
                 are not deleted.
    ===========  =====================================================

    To get x-y data points from plots from within a script, use
    :func:`DataCursor.getdata`. Enter the number of points or press
    't' to end blocking so the script will continue (see
    :func:`DataCursor.getdata`).

    Once the DataCursor is turned off, the annotations become
    draggable. Two notes on this:

        1. Due to a bug in matplotlib v1.4.3, dragging annotations
           connected to points with negative x or y coordinates
           doesn't work correctly. Newer versions work properly.
        2. Somehow, annotations sometimes become linked (moving one
           will move another). When that happens, try dragging
           a different annotation; this sometimes breaks the link.

    Interactively, the member functions :func:`DataCursor.on` and
    :func:`DataCursor.off` are used to turn the DataCursor on and
    off. These functions will update the internal state of the
    `DataCursor` to account for deleted or added items.
    :func:`DataCursor.getdata` calls :func:`DataCursor.on` internally.

    The following example plots some random data, calls
    :func:`DataCursor.getdata` to wait for the user to optionally
    select data points and then turn the DataCursor off (with
    keystroke 't'). It then prints the selected points::

        import matplotlib.pyplot as plt
        import numpy as np
        from pyyeti.datacursor import DC
        x = np.arange(500)/250
        y = np.random.rand(*x.shape)
        fig = plt.figure('demo')
        fig.clf()
        ax = fig.add_subplot(1, 1, 1)
        ax.plot(x, y)
        DC.getdata()
        # use DC.pause if you want to drag boxes around before
        # continuing:
        DC.pause()
        print('points are:', DC.xypoints)

    Settings can be changed after instantiation. Here is an example of
    defining a new format for the annotation (for both `form1` and
    `form2`). Only the line label is included in the annotation. The
    example also changes the permanent dot to a gray pentagon::

        import matplotlib.pyplot as plt
        import numpy as np
        from pyyeti.datacursor import DC

        def formnew(x, y, n, ind, ax, line):
            return f'{line.get_label()}'

        DC.form1 = DC.form2 = formnew
        DC.permdot = dict(s=130, color='black', alpha=0.4,
                          marker='p')

        plt.plot(np.random.randn(50), label='Gaussian')
        plt.plot(np.random.rand(50), label='Uniform')
        DC.on()

    For increased versatility, there are two optional functions the
    user can define that will be called when a point is added
    (left-click) and when a point is deleted (right-click). See
    :func:`DataCursor.addpt_func` and :func:`DataCursor.delpt_func`
    for more information on the call signatures. Here is a simple
    example that just prints statements to the screen::

        import matplotlib.pyplot as plt
        import numpy as np
        from pyyeti.datacursor import DC

        def addpt(x, y, n, ind, ax, line):
            print(f'You selected ({x}, {y})')

        def delpt(x, y, n, ind, ax, line):
            print(f'You deleted ({x}, {y})')

        DC.addpt_func(addpt)
        DC.delpt_func(delpt)

        plt.plot(np.random.randn(50))
        DC.on()
    """

    def __init__(
        self,
        ax=None,
        figs=None,
        hover=True,
        form1=form1,
        form2=form2,
        offsets=(-20, 20),
        bbox=dict(boxstyle="round,pad=0.5", fc="gray", alpha=0.5),
        arrowprops=dict(arrowstyle="->", connectionstyle="arc3,rad=0"),
        followdot=dict(s=130, color="green", alpha=0.7),
        permdot=dict(s=130, color="red", alpha=0.4),
    ):
        """
        Initialize the DataCursor.

        Parameters
        ----------
        ax : axes object(s) or None; optional
            Axes object or list of axes objects as created by
            :func:`matplotlib.pyplot.subplot` (for example). If None,
            all axes on all selected figures will be automatically
            included. Takes precedence over the `figs` input.
        figs : figure object(s) or None; optional
            Alternative to the `ax` input. If `ax` is not input,
            `figs` specifies a figure object or list of figure objects
            as created by :func:`matplotlib.pyplot.figure`. If None,
            all applicable figures will be automatically included.
        hover : bool; optional
            Sets the `hover` attribute.
        form1 : function; optional
            Sets the `form1` attribute.
        form2 : function; optional
            Sets the `form2` attribute.
        offsets : tuple; optional
            Sets the `offsets` attribute.
        bbox : dict; optional
            Sets the `bbox` attribute.
        arrowprops : dict; optional
            Sets the `arrowprops` attribute.
        followdot : dict; optional
            Sets the `followdot` attribute.
        permdot : dict; optional
            Sets the `permdot` attribute.
        """
        self._ax_input = ax
        self._figs_input = figs
        self.hover = hover
        self.form1 = form1
        self.form2 = form2
        self.offsets = offsets
        self.bbox = bbox
        self.arrowprops = arrowprops
        self.followdot = followdot
        self.permdot = permdot
        self._is_on = False
        self._in_loop = False
        self._max_points = -1
        self.on()
        self.addptfunc = None
        self.delptfunc = None

    def _init_all(self, errout=False):
        if self._ax_input is None:
            if self._figs_input is None:
                self._figs = [plt.figure(i) for i in plt.get_fignums()]
            else:
                self._figs = _ensure_iterable(self._figs_input)
            if len(self._figs) == 0:
                if errout:
                    raise RuntimeError("no figures; plot something first")
                else:
                    self._ax = []
                    return
            self._ax = [a for fig in self._figs for a in fig.get_axes()]
        else:
            self._ax = _ensure_iterable(self._ax_input)
            self._figs = [a.figure for a in self._ax]
        if len(self._ax) == 0:
            if errout:
                raise RuntimeError("no axes; plot something first")
            else:
                return
        maxlines = 0
        for a in self._ax:
            if len(a.lines) > maxlines:
                maxlines = len(a.lines)
        if maxlines == 0 and errout:
            raise RuntimeError("no lines; plot something first")

    def on(self, ax=-1, figs=-1, callbacks=True, reset=True):
        """
        Turns on and (re-)initializes the DataCursor for current
        figures.

        Parameters
        ----------
        ax : axes object(s) or None or -1; optional
            Axes object or list of axes objects as created by
            :func:`matplotlib.pyplot.subplot` (for example). If None,
            all axes on all selected figures will be automatically
            included. Takes precedence over the `figs` input. If -1,
            leave this setting as specified during instantiation.
        figs : figure object(s) or None or -1; optional
            Alternative to the `ax` input. If `ax` is not input,
            `figs` specifies a figure object or list of figure objects
            as created by :func:`matplotlib.pyplot.figure`. If None,
            all applicable figures will be automatically included. If
            -1, leave this setting as specified during instantiation.
        callbacks : bool or str; optional
            If False, call-backs are not turned on. If True, all call-
            backs are turned on. If 'key_only', only the key-press
            call-back is turned on (used for pausing; see
            :func:`pause`).
        reset : bool; optional
            If True, the `xypoints` and other data members are reset
            to empty lists. Otherwise, if `reset` is False, your new
            data will be appended on to your previous data.

        Returns
        -------
        None
        """
        if ax != -1:
            self._ax_input = ax
        if figs != -1:
            self._figs_input = figs
        self._init_all()
        if reset:
            self.xypoints = []
            self.inds = []
            self.lines = []
            self.linenums = []
            self.pts = []
            self.notes = []
        self._kid = {}  # key press event
        self._mid = {}  # motion event
        self._bid = {}  # button press event
        self._aid = {}  # axes leave event
        self._setup_annotations()
        if callbacks:
            for fig in self._figs:
                cvs = fig.canvas
                self._kid[fig] = cvs.mpl_connect("key_press_event", self._key)
                if callbacks != "key_only":
                    if self.hover:
                        self._mid[fig] = cvs.mpl_connect(
                            "motion_notify_event", self._follow
                        )
                    else:
                        self._mid[fig] = None
                    self._bid[fig] = cvs.mpl_connect("button_press_event", self._follow)
                    self._aid[fig] = cvs.mpl_connect("axes_leave_event", self._leave)
            self._is_on = True
        else:
            self._is_on = False

    def off(self, stop_blocking=True):
        """
        Turns off the DataCursor and optionally stops it from blocking

        Parameters
        ----------
        stop_blocking : bool; optional
            If True, have the data cursor stop blocking so Python can
            continue with whats next. Otherwise, if `stop_blocking` is
            False, Python will wait; this is probably only useful when
            the data cursor is controlled in a GUI environment.

        Notes
        -----
        Note that the keystroke 't' will also turn off the DataCursor;
        in that case, `stop_blocking` is True.
        """
        self._init_all()
        if self._is_on:
            for ax in self._ax:
                if ax in self._annotation:
                    self._annotation[ax].set_visible(False)
                    self._dot[ax].set_visible(False)
            for fig in self._figs:
                if fig in self._kid:
                    fig.canvas.mpl_disconnect(self._kid[fig])
                if self.hover and fig in self._mid and self._mid[fig] is not None:
                    fig.canvas.mpl_disconnect(self._mid[fig])
                if fig in self._bid:
                    fig.canvas.mpl_disconnect(self._bid[fig])
                    fig.canvas.mpl_disconnect(self._aid[fig])
            for an in self.notes:
                ax = an.axes
                if an in ax.texts:
                    # give annotation to figure so it's on top
                    ax.figure.texts.append(an)
                    ax.texts.pop(ax.texts.index(an))
                    # and make it draggable
                    an.draggable()
            self._is_on = False
        for fig in self._figs:
            fig.canvas.draw()
        if self._in_loop and stop_blocking:
            self._figs[0].canvas.stop_event_loop()
            self._in_loop = False

    def addpt_func(self, func):
        """
        Function to call on a left-click.

        Call signature is: ``func(x, y, n, ind, ax, line)``

        The parameters for the function are described below.

        Parameters
        ----------
        x, y : scalars
            The selected ``(x, y)`` data point
        n : integer
            The line number
        ind : integer
            The index of the data point within the plotted data
            vectors
        ax : axes object
            Axes handle
        line : line object
            Line handle
        """
        self.addptfunc = func

    def delpt_func(self, func):
        """
        Function to call on a right-click.

        Call signature is: ``func(x, y, n, ind, ax, line)``

        The parameters for the function are described below.

        Parameters
        ----------
        x, y : scalars
            The selected ``(x, y)`` data point
        n : integer
            The line number
        ind : integer
            The index of the data point within the plotted data
            vectors
        ax : axes object
            Axes handle
        line : line object
            Line handle or None; None if the line was deleted
        """
        self.delptfunc = func

    def _add_point(self, x, y, n, ind, ax, line):
        self.xypoints.append([x, y])
        self.inds.append(ind)
        self.lines.append(line)
        self.linenums.append(n)
        # if 3d, set visible to False (i don't know how to get proper
        # coordinates for the dot):
        vis = False if hasattr(ax, "get_proj") else True
        self.pts.append(ax.scatter(x, y, **self.permdot, visible=vis))
        self.notes.append(self._annotation[ax])
        # make a new annotation box so current one is static
        self._annotation[ax] = self._new_annotation(ax, (x, y))
        if self._in_loop and len(self.xypoints) == self._max_points:
            self.off()
        if self.addptfunc:
            self.addptfunc(x, y, n, ind, ax, line)

    def _del_point(self, ax=None, delete_line=False):
        """Deletes last saved point, if any."""
        if len(self.xypoints) == 0:
            return
        x, y = self.xypoints.pop()
        ind = self.inds.pop()
        h = self.lines.pop()
        if delete_line:
            # line may have been deleted already, so catch exception:
            try:
                h.remove()
            except ValueError:
                pass
            else:
                h = None
        n = self.linenums.pop()
        pt = self.pts.pop()
        pt.remove()
        na = self.notes.pop()
        na.remove()
        if self.delptfunc:
            self.delptfunc(x, y, n, ind, ax, h)

    def _get_ax(self, event):
        """
        Return axes for event, and possibly modifies event for xdata,
        ydata location.
        """
        if event.inaxes is None:
            return
        if event.inaxes in self._ax:
            ax = event.inaxes
        else:
            # for overlapping axes, only top axes gets the event, so
            # check others manually
            for ax in self._ax:
                pt = (event.x, event.y)
                if ax.contains_point(pt):
                    inv = ax.transData.inverted()
                    event.xdata, event.ydata = inv.transform_point(pt)
                    break
            else:
                return
        return ax

    def _key(self, event):
        """
        Processes 't' key press to turn the DataCursor off.
        """
        if not self._get_ax(event):
            return
        if event.key == "t" or event.key == "T":
            if self._is_on:
                self.off()
            else:
                self.on()
        elif event.key == "D":
            self._del_point(delete_line=True)
            event.canvas.draw()

    def _leave(self, event):
        """Event handler for when mouse leaves axes."""
        ax = self._get_ax(event)
        if not ax:
            return
        self._annotation[ax].set_visible(False)
        self._dot[ax].set_visible(False)
        event.canvas.draw()

    def _follow(self, event):
        """Event handler for when mouse is moved insided axes and for
        left and right clicks."""
        ax = self._get_ax(event)
        if not ax:
            return
        x, y = event.xdata, event.ydata
        x, y, n, ind, line = self._snap(ax, x, y)
        if x is None:
            return
        annotation = self._annotation[ax]
        dot = self._dot[ax]
        annotation.xy = x, y
        if len(ax.lines) > 1:
            annotation.set_text(self.form2(x, y, n, ind, ax, line))
        else:
            annotation.set_text(self.form1(x, y, n, ind, ax, line))
        dot.set_offsets((x, y))
        if event.name == "button_press_event":
            if event.button == 1:
                # dot.set_visible(True)
                annotation.set_visible(True)
                self._add_point(x, y, n, ind, ax, line)
            elif event.button == 3 and len(self.xypoints) > 0:
                self._del_point(ax=ax)
        elif self.hover:
            if not hasattr(ax, "get_proj"):
                dot.set_visible(True)
            annotation.set_visible(True)
        event.canvas.draw()

    def _new_annotation(self, ax, xy):
        return ax.annotate(
            "",
            xy=xy,
            ha="right",
            va="bottom",
            xytext=self.offsets,
            textcoords="offset points",
            bbox=self.bbox,
            arrowprops=self.arrowprops,
            visible=False,
        )

    def _setup_annotations(self):
        """Create the annotation boxes. The string value and the
        position will be set for each event."""
        self._annotation = {}
        self._dot = {}
        for ax in self._ax:
            if len(ax.lines) > 0:
                ln = ax.lines[0]
                xy = ln.get_xdata()[0], ln.get_ydata()[0]
            else:
                xy = 0, 0
            self._annotation[ax] = self._new_annotation(ax, xy)
            xl, yl = ax.get_xlim(), ax.get_ylim()
            self._dot[ax] = ax.scatter(xy[0], xy[1], **self.followdot, visible=False)
            ax.set_xlim(xl)
            ax.set_ylim(yl)

    def _snap(self, ax, x, y):
        """Return the value in self._points closest to x, y."""
        dmin = np.inf
        xyn = None, None, None, None, None
        for n, ln in enumerate(ax.lines):
            xdata = ln.get_xdata()
            ydata = ln.get_ydata()
            dx = (xdata - x) / np.diff(ax.get_xlim())[0]
            dy = (ydata - y) / np.diff(ax.get_ylim())[0]
            d = np.sqrt(dx ** 2.0 + dy ** 2.0)
            ind = np.argmin(d)
            if d[ind] < dmin:
                dmin = d[ind]
                xyn = xdata[ind], ydata[ind], n, ind, ln
        return xyn

    def _fake_getdata(self):
        """
        Available to bypass actual mouse interaction while running
        tests of routines that use :func:`getdata`. If this function
        returns anything other than None, :func:`getdata` is bypassed.
        """
        return

    def getdata(self, maxpoints=-1, msg='Select points, hit "t" inside axes when done'):
        """
        Suspend python while user selects points up to `maxpoints`.
        If `maxpoints` is < 0, the loop will last until user hits 't'
        inside the axes ('t' turns off the DataCursor). Deleted points
        are not counted.
        """
        # _fake_getdata is used for testing functions that need
        # getdata
        if self._fake_getdata() is None:
            if maxpoints == 0:
                return
            if maxpoints < 0 and msg:
                print(msg)
            self.on()
            self._in_loop = True
            self._max_points = maxpoints
            self._figs[0].canvas.start_event_loop(timeout=-1)

    def pause(self, msg='Pausing, hit "t" inside axes to continue'):
        """
        Suspend python so user can interact with plots (such as moving
        previously added annotations) before continuing. Hit 't'
        inside the axes to continue.
        """
        print(msg)
        self.on(callbacks="key_only", reset=False)
        self._in_loop = True
        self._figs[0].canvas.start_event_loop(timeout=-1)


# instantiate one object, meant for general use:
DC = DataCursor()



# -*- coding: utf-8 -*-
"""
pyYeti
======

pyYeti has tools mostly related to structural dynamics:

    * Solve matrix equations of motion in the time and
      frequency domains
    * Shock response spectrum (SRS)
    * Fatigue damage equivalent power spectral densities (PSD)
    * Hurty-Craig-Bampton model checks
    * Coupled loads analysis tools
    * Statistics tools for computing k-factors (for tolerance
      bounds and intervals) and for order statistics
    * A data-cursor for interacting with 2D x-y plots
    * Eigensystem Realization Algorithm
    * Force limiting analysis tools
    * Read/write Nastran output4 (.op4) files
    * Limited capability to read Nastran output2 (.op2) files
    * Rainflow cycle counting
    * Resample data with the Lanczos method
    * A "vectorized" writing module
    * Eigensolution with the subspace iteration method
    * Tools for working with the "nas2cam" Nastran DMAP
    * Other miscellaneous tools

"""

__version__ = "0.99.5"
